/*
 * Copyright (c) 2008 Microsoft Corporation.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id:$
 */

//
// IBAT: InfiniBand Address Translation
//
// Description:
//  Maps source & remote IP addresses (IPv4 and IPv6) to a path record.
//
//  The mapping requires two steps:
//      1. Mapping the remote IP address to the remote Ethernet MAC address
//      2. Retrieve the path given the remote Ethernet MAC address from IPoIB
//
//  The first step is accomplished as follows on Windows Server 2008:
//      1. Lookup the desired MAC from the OS using GetIpNetEntry2
//      2. If the remote IP isn't found, resolve the remote IP address
//      using ResolveIpNetEntry2
//
//  The first step is accomplished as follows on Windows Server 2003:
//      1. Retrieve the whole IP->MAC mapping table from the OS using
//      GetIpNetTable.
//      2. Walk the returned table looking for the destination IP to
//      find the destination Ethernet MAC address.
//      3. If the remote IP isn't found, resolve the remote IP address using
//      SendARP.
//
//  The second step is accomplished by asking IPoIB for the path
//  given the remote MAC.  IPoIB creates the path internally without going to
//  the SA.

#pragma warning( push, 3 )
#include <wtypes.h>

#include <stdlib.h>
#include <winioctl.h>
#pragma warning( pop )
#include "iba/ibat.h"
#include <iphlpapi.h>
#include "iba/ib_at_ioctl.h"

C_ASSERT( sizeof(IBAT_PATH_BLOB) == sizeof(ib_path_rec_t) );

namespace IBAT
{

    const IN6_ADDR x_DefaultGid = {0xFE,0x80,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

class H
{
public:
    H( HANDLE h = INVALID_HANDLE_VALUE ) : m_h( h ) {};
    ~H(){ if( m_h != INVALID_HANDLE_VALUE ) CloseHandle( m_h ); }

    H& operator =(HANDLE h){ CloseHandle( m_h ); m_h = h; }
    operator HANDLE() const { return m_h; }

private:
    HANDLE m_h;
};

#if WINVER >= 0x600
HRESULT
Resolve(
    __in const struct sockaddr* pSrcAddr,
    __in const struct sockaddr* pDestAddr,
    __out IBAT_PATH_BLOB* pPath
    )
{
    if( pSrcAddr->sa_family != pDestAddr->sa_family )
        return E_INVALIDARG;

    H hIbatDev = CreateFileW( IBAT_WIN32_NAME,
        MAXIMUM_ALLOWED,(FILE_SHARE_READ|FILE_SHARE_WRITE), NULL,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
    if( hIbatDev == INVALID_HANDLE_VALUE )
        return HRESULT_FROM_WIN32( GetLastError() );

    bool fLoopback;
    IOCTL_IBAT_IP_TO_PORT_IN port_in;
    port_in.Version = IBAT_IOCTL_VERSION;
    if( pSrcAddr->sa_family == AF_INET )
    {
        port_in.Address.IpVersion = 4;
        RtlCopyMemory(
            &port_in.Address.Address[12],
            &((struct sockaddr_in*)pSrcAddr)->sin_addr,
            sizeof( ((struct sockaddr_in*)pSrcAddr)->sin_addr ) );

        fLoopback = ((struct sockaddr_in*)pDestAddr)->sin_addr.s_addr ==
            ((struct sockaddr_in*)pSrcAddr)->sin_addr.s_addr;
    }
    else
    {
        port_in.Address.IpVersion = 6;
        RtlCopyMemory(
            port_in.Address.Address,
            &((struct sockaddr_in6*)pSrcAddr)->sin6_addr,
            sizeof(port_in.Address.Address) );
        fLoopback = IN6_ADDR_EQUAL(
            &((struct sockaddr_in6*)pDestAddr)->sin6_addr,
            &((struct sockaddr_in6*)pSrcAddr)->sin6_addr
            ) == TRUE;
    }

    IBAT_PORT_RECORD port_out;
    DWORD size;
    BOOL fSuccess = DeviceIoControl( hIbatDev, IOCTL_IBAT_IP_TO_PORT,
        &port_in, sizeof(port_in), &port_out, sizeof(port_out), &size, NULL );

    if( !fSuccess )
        return HRESULT_FROM_WIN32( GetLastError() );

    // Check for loopback.
    IOCTL_IBAT_MAC_TO_PATH_IN mac_in = {0};
    mac_in.Version = IBAT_IOCTL_VERSION;
    mac_in.PortGuid = port_out.PortGuid;
    if( !fLoopback )
    {
        NET_LUID luid;
        DWORD ret;
        do
        {
            DWORD iIf;
            ret = GetBestInterfaceEx( (struct sockaddr*)pSrcAddr, &iIf );
            if( ret != NO_ERROR )
                return HRESULT_FROM_WIN32( ret );

            // Interface indexes are not constant, so get the LUID mapping for the
            // returned interface for use in the rest of the function.
            ret = ConvertInterfaceIndexToLuid( iIf, &luid );

        } while( ret != NO_ERROR );

        SOCKADDR_INET src;
        MIB_IPNET_ROW2 net = {0};
        net.InterfaceLuid = luid;
        switch( pDestAddr->sa_family )
        {
        case AF_INET:
            net.Address.si_family = src.si_family = AF_INET;
            net.Address.Ipv4 = *(struct sockaddr_in*)pDestAddr;
            src.Ipv4 = *(struct sockaddr_in*)pSrcAddr;
            break;

        case AF_INET6:
            net.Address.si_family = src.si_family = AF_INET6;
            net.Address.Ipv6 = *(struct sockaddr_in6*)pDestAddr;
            src.Ipv6 = *(struct sockaddr_in6*)pSrcAddr;
            break;

        default:
            return E_INVALIDARG;
        }

        bool fRetry = true;
    retry:
        ret = GetIpNetEntry2( &net );
        if( ret == ERROR_NOT_FOUND )
        {
            net.State = NlnsUnreachable;
        }
        else if( ret != NO_ERROR )
        {
            return HRESULT_FROM_WIN32( ret );
        }

        switch( net.State )
        {
        default:
        case NlnsUnreachable:
            ret = ResolveIpNetEntry2( &net, &src );
            if( ret == ERROR_BAD_NET_NAME && fRetry )
            {
                fRetry = false;
                goto retry;
            }
            else if( ret != NO_ERROR )
            {
                return HRESULT_FROM_WIN32( ret );
            }
            break;

        case NlnsReachable:
        case NlnsPermanent:
            break;

        case NlnsIncomplete:
            return E_PENDING;
        }

        if( net.PhysicalAddressLength > 6 )
            return E_UNEXPECTED;

        RtlCopyMemory( mac_in.DestMac, net.PhysicalAddress, IBAT_MAC_LEN );
    }

    fSuccess = DeviceIoControl( hIbatDev, IOCTL_IBAT_MAC_TO_PATH,
        &mac_in, sizeof(mac_in), pPath, sizeof(*pPath), &size, NULL );
    if( !fSuccess )
        return HRESULT_FROM_WIN32( GetLastError() );

    return S_OK;
}
#else   // Back compatibility with Windows Server 2003


static HRESULT
GetDestMac(
    __in struct sockaddr_in* pDestAddr,
    __out BYTE* pDestMac
    )
{
    DWORD ret;

    MIB_IPNETTABLE* pTable = NULL;
    ULONG len = 0;
    do
    {
        ret = GetIpNetTable( pTable, &len, FALSE );
        if( ret != ERROR_INSUFFICIENT_BUFFER )
            break;

        if( pTable != NULL )
        {
            HeapFree( GetProcessHeap(), 0, pTable );
        }

        pTable = (MIB_IPNETTABLE*)HeapAlloc( GetProcessHeap(), 0, len );
    } while( ret == ERROR_INSUFFICIENT_BUFFER );

    if( ret != NO_ERROR )
    {
        if( pTable != NULL )
        {
            HeapFree( GetProcessHeap(), 0, pTable );
        }
        return HRESULT_FROM_WIN32( ret );
    }

    ret = ERROR_NOT_SUPPORTED;
    DWORD i;
    for( i = 0; i < pTable->dwNumEntries; i++ )
    {
        if( pTable->table[i].dwType == MIB_IPNET_TYPE_OTHER ||
            pTable->table[i].dwType == MIB_IPNET_TYPE_INVALID )
        {
            continue;
        }

        if( pTable->table[i].dwAddr !=
            ((struct sockaddr_in*)pDestAddr)->sin_addr.s_addr )
        {
            continue;
        }

        if( pTable->table[i].dwPhysAddrLen != IBAT_MAC_LEN )
        {
            continue;
        }

        RtlCopyMemory( pDestMac, pTable->table[i].bPhysAddr, IBAT_MAC_LEN );
        ret = S_OK;
        break;
    }
    HeapFree( GetProcessHeap(), 0, pTable );

    return HRESULT_FROM_WIN32( ret );
}

HRESULT
Resolve(
    __in const struct sockaddr* pSrcAddr,
    __in const struct sockaddr* pDestAddr,
    __out IBAT_PATH_BLOB* pPath
    )
{
    if( pDestAddr->sa_family != AF_INET )
        return E_NOTIMPL;

    H hIbatDev = CreateFileW( IBAT_WIN32_NAME,
        MAXIMUM_ALLOWED,(FILE_SHARE_READ|FILE_SHARE_WRITE), NULL,
        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
    if( hIbatDev == INVALID_HANDLE_VALUE )
        return HRESULT_FROM_WIN32( GetLastError() );

    IOCTL_IBAT_IP_TO_PORT_IN port_in;
    port_in.Version = IBAT_IOCTL_VERSION;
    port_in.Address.IpVersion = 4;
    RtlCopyMemory(
        &port_in.Address.Address[12],
        &((struct sockaddr_in*)pSrcAddr)->sin_addr,
        sizeof( ((struct sockaddr_in*)pSrcAddr)->sin_addr ) );

    IBAT_PORT_RECORD port_out;
    DWORD size;
    BOOL fSuccess = DeviceIoControl( hIbatDev, IOCTL_IBAT_IP_TO_PORT,
        &port_in, sizeof(port_in), &port_out, sizeof(port_out), &size, NULL );

    if( !fSuccess )
        return HRESULT_FROM_WIN32( GetLastError() );

    // Check for loopback.
    IOCTL_IBAT_MAC_TO_GID_IN mac_in = {0};
    mac_in.Version = IBAT_IOCTL_VERSION;
    mac_in.PortGuid = port_out.PortGuid;

    if( ((struct sockaddr_in*)pDestAddr)->sin_addr.s_addr !=
        ((struct sockaddr_in*)pSrcAddr)->sin_addr.s_addr )
    {
        HRESULT hr = GetDestMac( (struct sockaddr_in*)pDestAddr, mac_in.DestMac );
        if( FAILED( hr ) )
        {
            ULONG len = sizeof(mac_in.DestMac);
            DWORD ret = SendARP(
                ((struct sockaddr_in*)pDestAddr)->sin_addr.s_addr,
                ((struct sockaddr_in*)pSrcAddr)->sin_addr.s_addr,
                (ULONG*)mac_in.DestMac,
                &len
                );
            if( ret != NO_ERROR )
                return HRESULT_FROM_WIN32( ret );
        }
    }

    fSuccess = DeviceIoControl( hIbatDev, IOCTL_IBAT_MAC_TO_PATH,
        &mac_in, sizeof(mac_in), pPath, sizeof(*pPath), &size, NULL );
    if( !fSuccess )
        return HRESULT_FROM_WIN32( GetLastError() );

    return S_OK;
}

#endif // WINVER >= 0x600


HRESULT
ResolvePath(
    __in const struct sockaddr* pSrcAddr,
    __in const struct sockaddr* pDestAddr,
    __out IBAT_PATH_BLOB* pPath,
	__in DWORD Timeout)
{
	INT64 to;
	HRESULT hr;

	to = (Timeout == INFINITE) ? 0x7FFFFFFFFFFFFFFFL : (INT64) ((UINT64) Timeout);
	for (;;) {
		hr = Resolve(pSrcAddr, pDestAddr, pPath);
		if( hr != E_PENDING || to <= 0 )
			break;

		to -= 10;
		Sleep(10);
	};

	return hr;
}

} /* IBAT namespace */

extern "C"
{

HRESULT
IbatResolve(
    __in const struct sockaddr* pSrcAddr,
    __in const struct sockaddr* pDestAddr,
    __out IBAT_PATH_BLOB* pPath
    )
{
    return IBAT::Resolve( pSrcAddr, pDestAddr, pPath );
}

HRESULT
IbatResolvePath(
    __in const struct sockaddr* pSrcAddr,
    __in const struct sockaddr* pDestAddr,
    __out IBAT_PATH_BLOB* pPath,
	__in DWORD Timeout)
{
	return IBAT::ResolvePath(pSrcAddr, pDestAddr, pPath, Timeout);
}

} /* extern "C" */
