/*
 * Copyright (c) 2004, 2005 Topspin Communications.  All rights reserved.
 * Copyright (c) 2005 Sun Microsystems, Inc. All rights reserved.
 * Copyright (c) 2005 Mellanox Technologies. All rights reserved.
 * Copyright (c) 2006, 2007 Cisco Systems, Inc. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */


#include "mlx4.h"
#include "fw.h"
#include "icm.h"
#include "device.h"
#include "doorbell.h"
#include "complib\cl_thread.h"
#include <mlx4_debug.h>

#if defined(EVENT_TRACING)
#ifdef offsetof
#undef offsetof
#endif
#include "main.tmh"
#endif


static struct mlx4_profile default_profile = {
	1 << 17,	/* num_qp		*/
	1 << 4,		/* rdmarc_per_qp	*/
	1 << 16,	/* num_srq	*/
	1 << 16,	/* num_cq		*/
	1 << 13,	/* num_mcg	*/
	1 << 18,	/* num_mpt	*/ 
	1 << 20		/* num_mtt	*/
};

static void process_mod_param_profile(void)
{
	if (g.mod_num_qp)
		default_profile.num_qp = 1 << g.mod_num_qp;

	if (g.mod_rdmarc_per_qp)
		default_profile.rdmarc_per_qp = 1 << g.mod_rdmarc_per_qp;

	if (g.mod_num_srq)
		default_profile.num_srq = 1 << g.mod_num_srq;

	if (g.mod_num_cq)
		default_profile.num_cq = 1 << g.mod_num_cq;

	if (g.mod_num_mcg)
		default_profile.num_mcg = 1 << g.mod_num_mcg;

	if (g.mod_num_mpt)
		default_profile.num_mpt = 1 << g.mod_num_mpt;

	if (g.mod_num_mtt)
		default_profile.num_mtt = 1 << g.mod_num_mtt;
}

static struct pci_device_id 
mlx4_pci_table[] = {
	HCA(MELLANOX, SDR,		HERMON),
	HCA(MELLANOX, DDR,		HERMON),
	HCA(MELLANOX, DDR_G2,		HERMON),
	HCA(MELLANOX, DDR_G2_A,		HERMON),
	HCA(MELLANOX, QDR_G2,		HERMON),
	HCA(MELLANOX, QDR_G2_A,		HERMON),
	HCA(MELLANOX, ETH,		HERMON),
	HCA(MELLANOX, ETH_YATIR,		HERMON),
	HCA(MELLANOX, ETH_G2,		HERMON),
	HCA(MELLANOX, ETH_YATIR_G2, 	HERMON),
	HCA(MELLANOX, ETH_B0_G2,	HERMON),
	HCA(MELLANOX, ETH_B0_40Gb_G2,	HERMON),
	HCA(MELLANOX, BD,		LIVEFISH),
};
#define MLX4_PCI_TABLE_SIZE (sizeof(mlx4_pci_table)/sizeof(struct pci_device_id))


static int mlx4_check_port_params(struct mlx4_dev *dev,
				  enum mlx4_port_type *port_type)
{
	if (port_type[0] != port_type[1] &&
	    !(dev->caps.flags & MLX4_DEV_CAP_FLAG_DPDP)) {
		mlx4_err(dev, "Only same port types supported "
			      "on this HCA, aborting.\n");
		return -EINVAL;
	}
	if ((port_type[0] == MLX4_PORT_TYPE_ETH) &&
	    (port_type[1] == MLX4_PORT_TYPE_IB)) {
		mlx4_err(dev, "eth-ib configuration is not supported.\n");
		return -EINVAL;
	}
	return 0;
}

static void mlx4_str2port_type(WCHAR **port_str,
			       enum mlx4_port_type *port_type)
{
	int i;

	for (i = 0; i < MLX4_MAX_PORTS; i++) {
		if (!wcscmp(port_str[i], L"eth"))
			port_type[i] = MLX4_PORT_TYPE_ETH;
		else
			port_type[i] = MLX4_PORT_TYPE_IB;
	}
}

int mlx4_count_ib_ports(struct mlx4_dev *dev)
{
	int i;
	int count = 0;

	for (i = 0; i < dev->caps.num_ports; i++) {
		if (dev->caps.port_type[i+1] == MLX4_PORT_TYPE_IB) {
			count++;
		}
	}
	return count;
}

BOOLEAN mlx4_is_eth_port(struct mlx4_dev *dev, int port_number)
{
	if (dev->caps.port_type[port_number] == MLX4_PORT_TYPE_ETH) {
		return TRUE;
	}
	return FALSE;
}

BOOLEAN mlx4_is_ib_port(struct mlx4_dev *dev, int port_number)
{
	if (dev->caps.port_type[port_number] == MLX4_PORT_TYPE_IB){
        return TRUE;
    }
    return FALSE;        
}

BOOLEAN mlx4_is_enabled_port(struct mlx4_dev *dev, int port_number)
{
	if (dev->caps.port_state[port_number] == MLX4_PORT_ENABLED) {
		return TRUE;
	}
	return FALSE;
}

static void mlx4_set_port_mask(struct mlx4_dev *dev)
{
	int i;

	dev->caps.port_mask = 0;
	for (i = 1; i <= dev->caps.num_ports; ++i)
		if (dev->caps.port_type[i] == MLX4_PORT_TYPE_IB)
			dev->caps.port_mask |= 1 << (i - 1);
}

static int mlx4_dev_cap(struct mlx4_dev *dev, struct mlx4_dev_cap *dev_cap)
{
	int err;
	int i;
	int num_eth_ports = 0;
	enum mlx4_port_type port_type[MLX4_MAX_PORTS];
	struct mlx4_dev *mdev = dev;

	for (i = 0; i < MLX4_MAX_PORTS; i++) 
		port_type[i] = dev->dev_params.mod_port_type[i];

	err = mlx4_QUERY_DEV_CAP(dev, dev_cap);
	if (err) {
		mlx4_err(dev, "QUERY_DEV_CAP command failed, aborting.\n");
		return err;
	}

	if (dev_cap->min_page_sz > PAGE_SIZE) {
		mlx4_err(dev, "HCA minimum page size of %d bigger than "
			 "kernel PAGE_SIZE of %ld, aborting.\n",
			 dev_cap->min_page_sz, PAGE_SIZE);
		return -ENODEV;
	}
	if (dev_cap->num_ports > MLX4_MAX_PORTS) {
		mlx4_err(dev, "HCA has %d ports, but we only support %d, "
			 "aborting.\n",
			 dev_cap->num_ports, MLX4_MAX_PORTS);
		return -ENODEV;
	}

	if (dev_cap->uar_size > (int)pci_resource_len(dev->pdev, 2)) {
		mlx4_err(dev, "HCA reported UAR size of 0x%x bigger than "
			 "PCI resource 2 size of 0x%llx, aborting.\n",
			 dev_cap->uar_size,
			 (unsigned long long) pci_resource_len(dev->pdev, 2));
		return -ENODEV;
	}

	dev->caps.num_ports	     = dev_cap->num_ports;
	for (i = 1; i <= dev->caps.num_ports; ++i) {
		dev->caps.vl_cap[i]	    = dev_cap->max_vl[i];
		dev->caps.ib_mtu_cap[i]	    = dev_cap->ib_mtu[i];       
		dev->caps.gid_table_len[i]  = dev_cap->max_gids[i];
		dev->caps.pkey_table_len[i] = dev_cap->max_pkeys[i];
		dev->caps.port_width_cap[i] = (u8)dev_cap->max_port_width[i];
		dev->caps.eth_mtu_cap[i]    = dev_cap->eth_mtu[i];
		dev->caps.def_mac[i]	    = dev_cap->def_mac[i];        
	}

	dev->caps.num_uars	     = dev_cap->uar_size / PAGE_SIZE;
	dev->caps.local_ca_ack_delay = dev_cap->local_ca_ack_delay;
	dev->caps.bf_reg_size	     = dev_cap->bf_reg_size;
	dev->caps.bf_regs_per_page   = dev_cap->bf_regs_per_page;
	dev->caps.max_sq_sg	     = dev_cap->max_sq_sg;
	dev->caps.max_rq_sg	     = dev_cap->max_rq_sg;
	dev->caps.max_wqes	     = dev_cap->max_qp_sz;
	dev->caps.max_qp_init_rdma   = dev_cap->max_requester_per_qp;
	dev->caps.max_srq_wqes	     = dev_cap->max_srq_sz;
	dev->caps.max_srq_sge	     = dev_cap->max_rq_sg - 1;
	dev->caps.reserved_srqs	     = dev_cap->reserved_srqs;
	dev->caps.max_sq_desc_sz     = dev_cap->max_sq_desc_sz;
	dev->caps.max_rq_desc_sz     = dev_cap->max_rq_desc_sz;
	dev->caps.num_qp_per_mgm     = MLX4_QP_PER_MGM;
	/*
	 * Subtract 1 from the limit because we need to allocate a
	 * spare CQE so the HCA HW can tell the difference between an
	 * empty CQ and a full CQ.
	 */
	dev->caps.max_cqes	     = dev_cap->max_cq_sz - 1;
	dev->caps.reserved_cqs	     = dev_cap->reserved_cqs;
	dev->caps.reserved_eqs	     = dev_cap->reserved_eqs;
	dev->caps.reserved_mtts	     = DIV_ROUND_UP(dev_cap->reserved_mtts,
						    MLX4_MTT_ENTRY_PER_SEG);
	dev->caps.reserved_mrws	     = dev_cap->reserved_mrws;
	dev->caps.reserved_uars	     = dev_cap->reserved_uars;
	dev->caps.reserved_pds	     = dev_cap->reserved_pds;
	dev->caps.mtt_entry_sz	     = MLX4_MTT_ENTRY_PER_SEG * dev_cap->mtt_entry_sz;
	dev->caps.max_msg_sz         = dev_cap->max_msg_sz;
	dev->caps.page_size_cap	     = ~(u32) (dev_cap->min_page_sz - 1);
	dev->caps.flags		     = dev_cap->flags;
	dev->caps.stat_rate_support  = dev_cap->stat_rate_support;
	dev->caps.max_gso_sz         = dev_cap->max_gso_sz;

	dev->caps.log_num_macs 	= ilog2(roundup_pow_of_two
					(g.mod_num_mac + 1));
	dev->caps.log_num_vlans = ilog2(roundup_pow_of_two
					(g.mod_num_vlan + 2));
	dev->caps.log_num_prios = (g.mod_use_prio)? 3: 0;
	dev->caps.num_fc_exch = g.mod_num_fc_exch;

	err = mlx4_check_port_params(dev, port_type);
	if (err)
		return err;

	for (i = 1; i <= dev->caps.num_ports; ++i) {
		if (!dev_cap->supported_port_types[i]) {
			mlx4_warn(dev, "FW doesn't support Multi Protocol, "
				       "loading IB only\n");
			dev->caps.port_type[i] = MLX4_PORT_TYPE_IB;
			continue;
		}
		if (port_type[i-1] & dev_cap->supported_port_types[i])
			dev->caps.port_type[i] = port_type[i-1];
		else {
			if (dev_cap->supported_port_types[i] & MLX4_PORT_TYPE_IB)
				dev->caps.port_type[i] = MLX4_PORT_TYPE_IB;
			else
				dev->caps.port_type[i] = MLX4_PORT_TYPE_ETH;

			MLX4_PRINT_EV(TRACE_LEVEL_WARNING,MLX4_DBG_DRV ,
				("Port %d: Unsupported requested port type %#x. We'll use #%x from supported %#x! \n", 
				i, port_type[i-1], dev->caps.port_type[i], 
				(int)dev_cap->supported_port_types[i]));
			
			MLX4_PRINT_EV(TRACE_LEVEL_WARNING ,MLX4_DBG_DRV ,
				("Ven %x Dev %d Fw %d.%d.%d, IsBurnDevice %s\n", 
				(unsigned)dev->pdev->ven_id, (unsigned)dev->pdev->dev_id,
				(int) (dev->caps.fw_ver >> 32),
				(int) (dev->caps.fw_ver >> 16) & 0xffff, 
				(int) (dev->caps.fw_ver & 0xffff),
				mlx4_is_livefish(dev) ? "Y" : "N"
				));
		}
		if (dev->caps.log_num_macs > dev_cap->log_max_macs[i]) {
			dev->caps.log_num_macs = dev_cap->log_max_macs[i];
			mlx4_warn(dev, "Requested number of MACs is too much "
				       "for port %d, reducing to %d.\n",
					i, 1 << dev->caps.log_num_macs);
		}
		if (dev->caps.log_num_vlans > dev_cap->log_max_vlans[i]) {
			dev->caps.log_num_vlans = dev_cap->log_max_vlans[i];
			mlx4_warn(dev, "Requested number of VLANs is too much "
				       "for port %d, reducing to %d.\n",
					i, 1 << dev->caps.log_num_vlans);
		}
		if (dev->caps.port_type[i] == MLX4_PORT_TYPE_ETH)
			++num_eth_ports;
	}

	mlx4_set_port_mask(dev);

	dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FW] = dev_cap->reserved_qps;
	dev->caps.reserved_qps_cnt[MLX4_QP_REGION_ETH_ADDR] =
		dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FC_ADDR] =
		(1 << dev->caps.log_num_macs)*
		(1 << dev->caps.log_num_vlans)*
		(1 << dev->caps.log_num_prios)*
		num_eth_ports;
	dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FC_EXCH] = dev->caps.num_fc_exch;

	return 0;
}

static int __devinit mlx4_load_fw(struct mlx4_dev *dev)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	int err;

	priv->fw.fw_icm = mlx4_alloc_icm(dev, priv->fw.fw_pages,
					 GFP_HIGHUSER | __GFP_NOWARN, 0);
	if (!priv->fw.fw_icm) {
		mlx4_err(dev, "Couldn't allocate FW area, aborting.\n");
		return -ENOMEM;
	}

	err = mlx4_MAP_FA(dev, priv->fw.fw_icm);
	if (err) {
		mlx4_dbg(dev, "MAP_FA command failed, aborting.\n");
		WriteEventLogEntryData( dev->pdev->p_self_do, (ULONG)EVENT_MLX4_ERROR_MAP_FA, 0, 0, 1,
			L"%d", err );
		goto err_free;
	}

	err = mlx4_RUN_FW(dev);
	if (err) {
		mlx4_dbg(dev, "RUN_FW command failed, aborting.\n");
		WriteEventLogEntryData( dev->pdev->p_self_do, (ULONG)EVENT_MLX4_ERROR_RUN_FW, 0, 0, 1,
			L"%d", err );
		goto err_unmap_fa;
	}

	return 0;

err_unmap_fa:
	mlx4_UNMAP_FA(dev);

err_free:
	mlx4_free_icm(dev, priv->fw.fw_icm, 0);
	return err;
}

static int __devinit mlx4_init_cmpt_table(struct mlx4_dev *dev, u64 cmpt_base,
					  int cmpt_entry_sz)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	int err;

	err = mlx4_init_icm_table(dev, &priv->qp_table.cmpt_table,
				  cmpt_base +
				  ((u64) (MLX4_CMPT_TYPE_QP *
					  cmpt_entry_sz) << MLX4_CMPT_SHIFT),
				  cmpt_entry_sz, dev->caps.num_qps,
				  dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FW],
				  0, 0);
	if (err)
		goto err;

	err = mlx4_init_icm_table(dev, &priv->srq_table.cmpt_table,
				  cmpt_base +
				  ((u64) (MLX4_CMPT_TYPE_SRQ *
					  cmpt_entry_sz) << MLX4_CMPT_SHIFT),
				  cmpt_entry_sz, dev->caps.num_srqs,
				  dev->caps.reserved_srqs, 0, 0);
	if (err)
		goto err_qp;

	err = mlx4_init_icm_table(dev, &priv->cq_table.cmpt_table,
				  cmpt_base +
				  ((u64) (MLX4_CMPT_TYPE_CQ *
					  cmpt_entry_sz) << MLX4_CMPT_SHIFT),
				  cmpt_entry_sz, dev->caps.num_cqs,
				  dev->caps.reserved_cqs, 0, 0);
	if (err)
		goto err_srq;

	err = mlx4_init_icm_table(dev, &priv->eq_table.cmpt_table,
				  cmpt_base +
				  ((u64) (MLX4_CMPT_TYPE_EQ *
					  cmpt_entry_sz) << MLX4_CMPT_SHIFT),
				  cmpt_entry_sz,
				  roundup_pow_of_two(MLX4_NUM_EQ +
						     dev->caps.reserved_eqs),
				  MLX4_NUM_EQ + dev->caps.reserved_eqs, 0, 0);
	if (err)
		goto err_cq;

	return 0;

err_cq:
	mlx4_cleanup_icm_table(dev, &priv->cq_table.cmpt_table);

err_srq:
	mlx4_cleanup_icm_table(dev, &priv->srq_table.cmpt_table);

err_qp:
	mlx4_cleanup_icm_table(dev, &priv->qp_table.cmpt_table);

err:
	return err;
}

static int mlx4_init_icm(struct mlx4_dev *dev, struct mlx4_dev_cap *dev_cap,
			 struct mlx4_init_hca_param *init_hca, u64 icm_size)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	u64 aux_pages;
	int err;

	err = mlx4_SET_ICM_SIZE(dev, icm_size, &aux_pages);
	if (err) {
		mlx4_err(dev, "SET_ICM_SIZE command failed, aborting.\n");
		return err;
	}

	mlx4_dbg(dev, "%lld KB of HCA context requires %lld KB aux memory.\n",
		 (unsigned long long) icm_size >> 10,
		 (unsigned long long) aux_pages << 2);

	priv->fw.aux_icm = mlx4_alloc_icm(dev, (int)aux_pages,
					  GFP_HIGHUSER | __GFP_NOWARN, 0);
	if (!priv->fw.aux_icm) {
		mlx4_err(dev, "Couldn't allocate aux memory, aborting.\n");
		return -ENOMEM;
	}

	err = mlx4_MAP_ICM_AUX(dev, priv->fw.aux_icm);
	if (err) {
		mlx4_err(dev, "MAP_ICM_AUX command failed, aborting.\n");
		goto err_free_aux;
	}

	err = mlx4_init_cmpt_table(dev, init_hca->cmpt_base, dev_cap->cmpt_entry_sz);
	if (err) {
		mlx4_err(dev, "Failed to map cMPT context memory, aborting.\n");
		goto err_unmap_aux;
	}

	err = mlx4_init_icm_table(dev, &priv->eq_table.table,
		init_hca->eqc_base, dev_cap->eqc_entry_sz,
		dev->caps.num_eqs, dev->caps.num_eqs, 0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map EQ context memory, aborting.\n");
		goto err_unmap_cmpt;
	}

	/*
	 * Reserved MTT entries must be aligned up to a cacheline
	 * boundary, since the FW will write to them, while the driver
	 * writes to all other MTT entries. (The variable
	 * dev->caps.mtt_entry_sz below is really the MTT segment
	 * size, not the raw entry size)
	 */
	dev->caps.reserved_mtts =
		ALIGN(dev->caps.reserved_mtts * dev->caps.mtt_entry_sz,
		      dma_get_cache_alignment()) / dev->caps.mtt_entry_sz;
	if ( dev->pdev->p_self_do->AlignmentRequirement + 1 != dma_get_cache_alignment()) {
		mlx4_dbg(dev, "Cache-line size %d, recommended value %d.\n",
			dev->pdev->p_self_do->AlignmentRequirement + 1,
			dma_get_cache_alignment() );
	}

	err = mlx4_init_icm_table(dev, &priv->mr_table.mtt_table,
				  init_hca->mtt_base,
				  dev->caps.mtt_entry_sz,
				  dev->caps.num_mtt_segs,
				  dev->caps.reserved_mtts, 1, 0);
	if (err) {
		mlx4_err(dev, "Failed to map MTT context memory, aborting.\n");
		goto err_unmap_eq;
	}

	err = mlx4_init_icm_table(dev, &priv->mr_table.dmpt_table,
				  init_hca->dmpt_base,
				  dev_cap->dmpt_entry_sz,
				  dev->caps.num_mpts,
				  dev->caps.reserved_mrws, 1, 1);
	if (err) {
		mlx4_err(dev, "Failed to map dMPT context memory, aborting.\n");
		goto err_unmap_mtt;
	}

	err = mlx4_init_icm_table(dev, &priv->qp_table.qp_table,
				  init_hca->qpc_base,
				  dev_cap->qpc_entry_sz,
				  dev->caps.num_qps,
				  dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FW],
				  0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map QP context memory, aborting.\n");
		goto err_unmap_dmpt;
	}

	err = mlx4_init_icm_table(dev, &priv->qp_table.auxc_table,
				  init_hca->auxc_base,
				  dev_cap->aux_entry_sz,
				  dev->caps.num_qps,
				  dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FW],
				  0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map AUXC context memory, aborting.\n");
		goto err_unmap_qp;
	}

	err = mlx4_init_icm_table(dev, &priv->qp_table.altc_table,
				  init_hca->altc_base,
				  dev_cap->altc_entry_sz,
				  dev->caps.num_qps,
				  dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FW],
				  0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map ALTC context memory, aborting.\n");
		goto err_unmap_auxc;
	}

	err = mlx4_init_icm_table(dev, &priv->qp_table.rdmarc_table,
				  init_hca->rdmarc_base,
				  dev_cap->rdmarc_entry_sz << priv->qp_table.rdmarc_shift,
				  dev->caps.num_qps,
				  dev->caps.reserved_qps_cnt[MLX4_QP_REGION_FW],
				  0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map RDMARC context memory, aborting\n");
		goto err_unmap_altc;
	}

	err = mlx4_init_icm_table(dev, &priv->cq_table.table,
				  init_hca->cqc_base,
				  dev_cap->cqc_entry_sz,
				  dev->caps.num_cqs,
				  dev->caps.reserved_cqs, 0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map CQ context memory, aborting.\n");
		goto err_unmap_rdmarc;
	}

	err = mlx4_init_icm_table(dev, &priv->srq_table.table,
				  init_hca->srqc_base,
				  dev_cap->srq_entry_sz,
				  dev->caps.num_srqs,
				  dev->caps.reserved_srqs, 0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map SRQ context memory, aborting.\n");
		goto err_unmap_cq;
	}

	/*
	 * It's not strictly required, but for simplicity just map the
	 * whole multicast group table now.  The table isn't very big
	 * and it's a lot easier than trying to track ref counts.
	 */
	err = mlx4_init_icm_table(dev, &priv->mcg_table.table,
				  init_hca->mc_base, MLX4_MGM_ENTRY_SIZE,
				  dev->caps.num_mgms + dev->caps.num_amgms,
				  dev->caps.num_mgms + dev->caps.num_amgms,
				  0, 0);
	if (err) {
		mlx4_err(dev, "Failed to map MCG context memory, aborting.\n");
		goto err_unmap_srq;
	}

	return 0;

err_unmap_srq:
	mlx4_cleanup_icm_table(dev, &priv->srq_table.table);

err_unmap_cq:
	mlx4_cleanup_icm_table(dev, &priv->cq_table.table);

err_unmap_rdmarc:
	mlx4_cleanup_icm_table(dev, &priv->qp_table.rdmarc_table);

err_unmap_altc:
	mlx4_cleanup_icm_table(dev, &priv->qp_table.altc_table);

err_unmap_auxc:
	mlx4_cleanup_icm_table(dev, &priv->qp_table.auxc_table);

err_unmap_qp:
	mlx4_cleanup_icm_table(dev, &priv->qp_table.qp_table);

err_unmap_dmpt:
	mlx4_cleanup_icm_table(dev, &priv->mr_table.dmpt_table);

err_unmap_mtt:
	mlx4_cleanup_icm_table(dev, &priv->mr_table.mtt_table);

err_unmap_eq:
	mlx4_cleanup_icm_table(dev, &priv->eq_table.table);

err_unmap_cmpt:
	mlx4_cleanup_icm_table(dev, &priv->eq_table.cmpt_table);
	mlx4_cleanup_icm_table(dev, &priv->cq_table.cmpt_table);
	mlx4_cleanup_icm_table(dev, &priv->srq_table.cmpt_table);
	mlx4_cleanup_icm_table(dev, &priv->qp_table.cmpt_table);

err_unmap_aux:
	mlx4_UNMAP_ICM_AUX(dev);

err_free_aux:
	mlx4_free_icm(dev, priv->fw.aux_icm, 0);

	return err;
}

static void mlx4_free_icms(struct mlx4_dev *dev)
{
	struct mlx4_priv *priv = mlx4_priv(dev);

	mlx4_cleanup_icm_table(dev, &priv->mcg_table.table);
	mlx4_cleanup_icm_table(dev, &priv->srq_table.table);
	mlx4_cleanup_icm_table(dev, &priv->cq_table.table);
	mlx4_cleanup_icm_table(dev, &priv->qp_table.rdmarc_table);
	mlx4_cleanup_icm_table(dev, &priv->qp_table.altc_table);
	mlx4_cleanup_icm_table(dev, &priv->qp_table.auxc_table);
	mlx4_cleanup_icm_table(dev, &priv->qp_table.qp_table);
	mlx4_cleanup_icm_table(dev, &priv->mr_table.dmpt_table);
	mlx4_cleanup_icm_table(dev, &priv->mr_table.mtt_table);
	mlx4_cleanup_icm_table(dev, &priv->eq_table.table);	
	mlx4_cleanup_icm_table(dev, &priv->eq_table.cmpt_table);
	mlx4_cleanup_icm_table(dev, &priv->cq_table.cmpt_table);
	mlx4_cleanup_icm_table(dev, &priv->srq_table.cmpt_table);
	mlx4_cleanup_icm_table(dev, &priv->qp_table.cmpt_table);

	mlx4_UNMAP_ICM_AUX(dev);
	mlx4_free_icm(dev, priv->fw.aux_icm, 0);
    priv->fw.aux_icm = NULL;
}

static void mlx4_close_hca(struct mlx4_dev *dev)
{
	mlx4_CLOSE_HCA(dev, 0);
	mlx4_free_icms(dev);
	mlx4_UNMAP_FA(dev);
	mlx4_free_icm(dev, mlx4_priv(dev)->fw.fw_icm, 0);
}

static int mlx4_init_hca(struct mlx4_dev *dev)
{
	struct mlx4_priv	  *priv = mlx4_priv(dev);
	struct mlx4_adapter	   adapter;
	struct mlx4_dev_cap	   dev_cap;
	struct mlx4_profile	   profile;
	struct mlx4_init_hca_param init_hca;
	u64 icm_size;
	int err;

	err = mlx4_QUERY_FW(dev);
	if (err) {
		if (err == -EACCES) {
			static int print_it = 1;
			if (print_it-- > 0) {
				mlx4_info(dev, "Function disabled, please upgrade to multi function driver.\n");
				WriteEventLogEntryData( dev->pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_QUERY_FW, 0, 0, 0 );
			}
		}
		else {
			mlx4_dbg(dev, "QUERY_FW command failed, aborting.\n");
			WriteEventLogEntryData( dev->pdev->p_self_do, (ULONG)EVENT_MLX4_ERROR_QUERY_FW, 0, 0, 1,
				L"%d", err );
		}
		return err;
	}

	err = mlx4_load_fw(dev);
	if (err) {
		mlx4_dbg(dev, "Failed to start FW, aborting.\n");
		return err;
	}

	err = mlx4_dev_cap(dev, &dev_cap);
	if (err) {
		mlx4_dbg(dev, "QUERY_DEV_CAP command failed, aborting.\n");
		WriteEventLogEntryData( dev->pdev->p_self_do, (ULONG)EVENT_MLX4_ERROR_QUERY_DEV_CAP, 0, 0, 1,
			L"%d", err );
		goto err_stop_fw;
	}

    //
    // Initilize the port state. It's a new command that is supported only in FW 2.6.1280. 
    //If running on earlier FW version the command can fail. Ignore the error code returned
    //
    mlx4_port_state(dev);

	process_mod_param_profile();
	profile = default_profile;

	icm_size = mlx4_make_profile(dev, &profile, &dev_cap, &init_hca);
	if ((long long) icm_size < 0) {
		err = (int)icm_size;
		goto err_stop_fw;
	}

	init_hca.log_uar_sz = (u8)ilog2(dev->caps.num_uars);

	err = mlx4_init_icm(dev, &dev_cap, &init_hca, icm_size);
	if (err)
		goto err_stop_fw;

	err = mlx4_INIT_HCA(dev, &init_hca);
	if (err) {
		mlx4_err(dev, "INIT_HCA command failed, aborting.\n");
		goto err_free_icm;
	}

	err = mlx4_QUERY_ADAPTER(dev, &adapter);
	if (err) {
		mlx4_err(dev, "QUERY_ADAPTER command failed, aborting.\n");
		WriteEventLogEntryData( dev->pdev->p_self_do, (ULONG)EVENT_MLX4_ERROR_QUERY_ADAPTER, 0, 0, 1,
			L"%d", err );
		goto err_close;
	}

	priv->eq_table.inta_pin = adapter.inta_pin;
	memcpy(dev->board_id, adapter.board_id, sizeof dev->board_id);

	return 0;

err_close:
	mlx4_CLOSE_HCA(dev,0);

err_free_icm:
	mlx4_free_icms(dev);

err_stop_fw:
	mlx4_UNMAP_FA(dev);
	mlx4_free_icm(dev, priv->fw.fw_icm, 0);

	return err;
}

static int mlx4_setup_hca(struct mlx4_dev *dev)
{
	struct mlx4_priv *priv = mlx4_priv(dev);
	int err;
	u8 port;

	err = mlx4_init_uar_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "user access region table, aborting.\n");
		return err;
	}

	err = mlx4_uar_alloc(dev, &priv->driver_uar);
	if (err) {
		mlx4_err(dev, "Failed to allocate driver access region, "
			 "aborting.\n");
		goto err_uar_table_free;
	}

	priv->kar = ioremap((u64)priv->driver_uar.pfn << PAGE_SHIFT, PAGE_SIZE, MmNonCached);
	if (!priv->kar) {
		mlx4_err(dev, "Couldn't map kernel access region, "
			 "aborting.\n");
		err = -ENOMEM;
		goto err_uar_free;
	}

	err = mlx4_init_pd_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "protection domain table, aborting.\n");
		goto err_kar_unmap;
	}

	err = mlx4_init_mr_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "memory region table, aborting.\n");
		goto err_pd_table_free;
	}

	
	err = mlx4_init_eq_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "event queue table, aborting.\n");
		goto err_mr_table_free;
	}

	err = mlx4_cmd_use_events(dev);
	if (err) {
		mlx4_err(dev, "Failed to switch to event-driven "
			 "firmware commands, aborting.\n");
		goto err_eq_table_free;
	}

	err = mlx4_NOP(dev);
	if (err) {
		if (dev->flags & MLX4_FLAG_MSI_X) {
			mlx4_warn(dev, "NOP command failed to generate MSI-X "
				  "interrupt IRQ %d.\n",
				  priv->eq_table.eq[MLX4_EQ_ASYNC].irq);
			mlx4_warn(dev, "Trying again without MSI-X.\n");
		} else {
			mlx4_err(dev, "NOP command failed to generate interrupt "
				 "(IRQ %d), aborting.\n",
				 priv->eq_table.eq[MLX4_EQ_ASYNC].irq);
			mlx4_err(dev, "BIOS or ACPI interrupt routing problem?\n");
		}

		goto err_cmd_poll;
	}

	mlx4_dbg(dev, "NOP command IRQ test passed\n");

	err = mlx4_init_cq_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "completion queue table, aborting.\n");
		goto err_cmd_poll;
	}

	err = mlx4_init_srq_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "shared receive queue table, aborting.\n");
		goto err_cq_table_free;
	}

	err = mlx4_init_qp_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "queue pair table, aborting.\n");
		goto err_srq_table_free;
	}

	err = mlx4_init_mcg_table(dev);
	if (err) {
		mlx4_err(dev, "Failed to initialize "
			 "multicast group table, aborting.\n");
		goto err_qp_table_free;
	}
	for (port = 1; port <= dev->caps.num_ports; port++) {
		err = mlx4_SET_PORT(dev, port,0 ,0);
		if (err) {
			mlx4_err(dev, "Failed to set port %d, aborting\n",
				 port);
			goto err_mcg_table_free;
		}
	}

	for (port = 0; port < dev->caps.num_ports; port++) {
		mlx4_init_mac_table(dev, port);
		mlx4_init_vlan_table(dev, port);
	}

	return 0;
err_mcg_table_free:
	mlx4_cleanup_mcg_table(dev);

err_qp_table_free:
	mlx4_cleanup_qp_table(dev);

err_srq_table_free:
	mlx4_cleanup_srq_table(dev);

err_cq_table_free:
	mlx4_cleanup_cq_table(dev);

err_cmd_poll:
	mlx4_cmd_use_polling(dev);

err_eq_table_free:
	mlx4_cleanup_eq_table(dev);

err_mr_table_free:
	mlx4_cleanup_mr_table(dev);

err_pd_table_free:
	mlx4_cleanup_pd_table(dev);

err_kar_unmap:
	iounmap(priv->kar,PAGE_SIZE);

err_uar_free:
	mlx4_uar_free(dev, &priv->driver_uar);

err_uar_table_free:
	mlx4_cleanup_uar_table(dev);
	return err;
}

static void __devinit mlx4_enable_msi_x(struct mlx4_dev *dev)
{
	int n_cpus = num_possible_cpus();

	/* decide on the mode */
	if (dev->pdev->n_msi_vectors_alloc >= (n_cpus+2))
		dev->flags |= MLX4_FLAG_MSI_X;

	MLX4_PRINT(TRACE_LEVEL_WARNING ,MLX4_DBG_LOW ,
		("Interrupt mode '%s', n_vectors %d, n_cpus %d\n",
		(dev->flags & MLX4_FLAG_MSI_X) ? "MSI-X" : "Legacy", 
		dev->pdev->n_msi_vectors_alloc, n_cpus));

}


static struct pci_device_id * mlx4_find_pci_dev(USHORT ven_id, USHORT dev_id)
{
	struct pci_device_id *p_id = mlx4_pci_table;
	int i;

	// find p_id (appropriate line in mlx4_pci_table)
	for (i = 0; i < MLX4_PCI_TABLE_SIZE; ++i, ++p_id) {
		if (p_id->device == dev_id && p_id->vendor ==  ven_id)
			return p_id;
	}
	return NULL;
}


int mlx4_init_one(struct pci_dev *pdev, struct mlx4_dev_params *dev_params)
{
	struct pci_device_id *id;
	struct mlx4_priv *priv;
	struct mlx4_dev *dev;
	int err;
	NTSTATUS status;
	int i;

#ifdef FORCE_LIVEFISH
		if (pdev)
			goto err;
#endif

	/* we are going to recreate device anyway */
	pdev->dev = NULL;
	pdev->ib_dev = NULL;
	
	/* find the type of device */
	id = mlx4_find_pci_dev(pdev->ven_id, pdev->dev_id);
	if (id == NULL) {
		err = -ENOSYS;
		goto err;
	}

	/*
	 * Check for BARs.  We expect 0: 1MB, 2: 8MB, 4: DDR (may not
	 * be present)
	 */
	if (pci_resource_len(pdev, 0) != 1 << 20) {
		MLX4_PRINT(TRACE_LEVEL_ERROR ,MLX4_DBG_LOW ,
			("Missing DCS, aborting.\n"));
		err = -ENODEV;
		goto err;
	}
	if (!pci_resource_len(pdev, 1)) {
		MLX4_PRINT(TRACE_LEVEL_ERROR ,MLX4_DBG_LOW ,
			("Missing UAR, aborting.\n"));
		err = -ENODEV;
		goto err;
	}

run_as_livefish:
	/* allocate mlx4_priv structure */
	priv = kzalloc(sizeof *priv, GFP_KERNEL);
	if (!priv) {
		MLX4_PRINT(TRACE_LEVEL_INFORMATION ,MLX4_DBG_LOW ,
			("Device struct alloc failed, aborting.\n"));
		err = -ENOMEM;
		goto end;
	}
	/* must be here for livefish */
	INIT_LIST_HEAD(&priv->ctx_list);
	spin_lock_init(&priv->ctx_lock);

	INIT_LIST_HEAD(&priv->pgdir_list);
	mutex_init(&priv->pgdir_mutex);

	/* deal with livefish, if any */
	dev       = &priv->dev;
	dev->signature = MLX4_DEV_SIGNATURE;
	dev->pdev = pdev;
	pdev->dev = dev;
	if (id->driver_data == LIVEFISH)
		dev->flags |= MLX4_FLAG_LIVEFISH;
	if (mlx4_is_livefish(dev)) {
		err = mlx4_register_device(dev);
		if (err) {
			MLX4_PRINT(TRACE_LEVEL_INFORMATION ,MLX4_DBG_LOW ,
				("mlx4_register_device for livefish failed, return with error.\n"));
			WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_ERROR_LIFEFISH_FAIL, 
				0, errno_to_ntstatus(err), 0 ); 
			pdev->dev = NULL;
			kfree(priv);
		} 
		else {
    		MLX4_PRINT(TRACE_LEVEL_ERROR ,MLX4_DBG_LOW ,
    			("MLX4_BUS started in \"livefish\" mode !!!.\n"));
			WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_ERROR_LIFEFISH_OK, 
				0, errno_to_ntstatus(err), 0 ); 
		}
		goto end;
	}

	for (i = 0; i < MLX4_MAX_PORTS; i++) 
		dev->dev_params.mod_port_type[i] = dev_params->mod_port_type[i];

	/*
	 * Now reset the HCA before we touch the PCI capabilities or
	 * attempt a firmware command, since a boot ROM may have left
	 * the HCA in an undefined state.
	 */
	status = mlx4_reset(dev);
	if ( !NT_SUCCESS( status ) ) {
		mlx4_err(dev, "Failed to reset HCA, aborting.(status %#x)\n", status);
		err = -EFAULT;
		goto err_free_dev;
	}

	if (mlx4_cmd_init(dev)) {
		mlx4_err(dev, "Failed to init command interface, aborting.\n");
		goto err_free_dev;
	}

	err = mlx4_init_hca(dev);
	if (err) {
		if (err == -EACCES) 
			dev->flags |= MLX4_FLAG_NOT_PRIME;
		goto err_cmd;
	}

	mlx4_enable_msi_x(dev);

	err = mlx4_setup_hca(dev);
	if (err == -EBUSY && (dev->flags & MLX4_FLAG_MSI_X)) {
		dev->flags &= ~MLX4_FLAG_MSI_X;
		err = mlx4_setup_hca(dev);
	}

	if (err)
		goto err_close;

	err = mlx4_register_device(dev);
	if (err)
		goto err_cleanup;

	mlx4_dbg(dev, "MLX4_BUS: NET device (dev_id=%d) is INITIALIZED ! \n", (int)pdev->dev_id);
	return 0;

err_cleanup:
	mlx4_cleanup_mcg_table(dev);
	mlx4_cleanup_qp_table(dev);
	mlx4_cleanup_srq_table(dev);
	mlx4_cleanup_cq_table(dev);
	mlx4_cmd_use_polling(dev);
	mlx4_cleanup_eq_table(dev);
	mlx4_cleanup_mr_table(dev);
	mlx4_cleanup_pd_table(dev);
	mlx4_cleanup_uar_table(dev);

err_close:
	mlx4_close_hca(dev);

err_cmd:
	mlx4_cmd_cleanup(dev);

err_free_dev:
	pdev->dev = NULL;
	kfree(priv);

err:
	/* we failed device initialization - try to simulate "livefish" device to facilitate using FW burning tools */
	id = mlx4_find_pci_dev(pdev->ven_id, DEVID_HERMON_BD);
	if (id == NULL) {
		err = -ENOSYS;
		goto end;
	}
	goto run_as_livefish;

end:	
	return err;
}

void mlx4_remove_one(struct pci_dev *pdev, int reset)
{
	struct mlx4_dev  *dev  = pdev->dev;
	struct mlx4_priv *priv = mlx4_priv(dev);
	int p;

	if (dev) {
		mlx4_unregister_device(dev);
		if (mlx4_is_livefish(dev))
			goto done;

		for (p = 1; p <= dev->caps.num_ports; ++p)
			mlx4_CLOSE_PORT(dev, p);

		mlx4_cleanup_mcg_table(dev);
		mlx4_cleanup_qp_table(dev);
		mlx4_cleanup_srq_table(dev);
		mlx4_cleanup_cq_table(dev);
		mlx4_cmd_use_polling(dev);
		mlx4_cleanup_eq_table(dev);
		mlx4_cleanup_mr_table(dev);
		mlx4_cleanup_pd_table(dev);

		iounmap(priv->kar,PAGE_SIZE);
		mlx4_uar_free(dev, &priv->driver_uar);
		mlx4_cleanup_uar_table(dev);
		mlx4_close_hca(dev);
		mlx4_cmd_cleanup(dev);

		if (reset && mlx4_reset(dev))
			mlx4_err(dev, "Failed to reset HCA\n");
		mlx4_dbg(dev, "MLX4_BUS: NET device (dev_id=%d) is REMOVED ! \n", (int)pdev->dev_id);
		pdev->dev = NULL;
done:
		kfree(priv);
	}
}

int mlx4_restart_one(struct pci_dev *pdev)
{
	struct mlx4_dev_params dev_params;
	mlx4_copy_dev_params(&dev_params, &pdev->dev->dev_params);

	mlx4_remove_one(pdev, FALSE);
	return mlx4_init_one(pdev, &dev_params);
}

void mlx4_net_init()
{
	mlx4_intf_init();
}



