#pragma once

#include "iobuf.h"
#include "complib\cl_debug.h"

////////////////////////////////////////////////////////
//
// CONSTANTS
//
////////////////////////////////////////////////////////

#define MT_TAG_ATOMIC		'MOTA'
#define MT_TAG_KERNEL		'LNRK'
#define MT_TAG_HIGH			'HGIH'
#define MT_TAG_PCIPOOL		'PICP'
#define MT_TAG_IOMAP		'PAMI'

////////////////////////////////////////////////////////
//
// SUBSTITUTIONS
//
////////////////////////////////////////////////////////


////////////////////////////////////////////////////////
//
// MACROS
//
////////////////////////////////////////////////////////

#define PAGE_MASK						(~(PAGE_SIZE-1))


 ////////////////////////////////////////////////////////
 //
 // Helper functions
 //
 ////////////////////////////////////////////////////////

// returns log of number of pages, i.e
// for size <= 4096 ==> 0
// for size <= 8192 ==> 1
static inline int get_order(unsigned long size)
{
	int order;

	size = (size-1) >> (PAGE_SHIFT-1);
	order = -1;
	do {
		size >>= 1;
		order++;
	} while (size);
	return order;
}

static inline unsigned long roundup_pow_of_two(unsigned long x)
{
	return (1UL << fls(x - 1));
}



////////////////////////////////////////////////////////
//
// SYSTEM MEMORY
//
////////////////////////////////////////////////////////

typedef enum _gfp {
	__GFP_NOWARN		= 0,	/* Suppress page allocation failure warning */
	__GFP_HIGHMEM		= 0,	/* high memory */
	GFP_ATOMIC			= 1,	/* can't wait (i.e. DPC or higher) */
	GFP_KERNEL			= 2,	/* can wait (npaged) */
	GFP_HIGHUSER 		= 4	/* GFP_KERNEL, that can be in HIGH memory */
}
gfp_t;

struct vm_area_struct {
	void *	ptr;
};

static inline void * kmalloc( SIZE_T bsize, gfp_t gfp_mask)
{
	void *ptr;
	ASSERT( KeGetCurrentIrql() <= DISPATCH_LEVEL);
	ASSERT(bsize);
	switch (gfp_mask) {
		case GFP_ATOMIC:
			ptr = ExAllocatePoolWithTag( NonPagedPool, bsize, MT_TAG_ATOMIC );
			break;
		case GFP_KERNEL:
			ptr = ExAllocatePoolWithTag( NonPagedPool, bsize, MT_TAG_KERNEL );
			break;
		case GFP_HIGHUSER:
			ptr = ExAllocatePoolWithTag( NonPagedPool, bsize, MT_TAG_HIGH );
			break;
		default:
			cl_dbg_out("kmalloc: unsupported flag %d\n", gfp_mask);
			ptr = NULL;
			break;
	}
	return ptr;
}

static inline void * kzalloc( SIZE_T bsize, gfp_t gfp_mask)
{
	void* va = kmalloc(bsize, gfp_mask);
	if (va)
		RtlZeroMemory(va, bsize);
	return va;
}

static inline void *kcalloc(size_t n, size_t size, gfp_t flags)
{
	if (n != 0 && size > ULONG_MAX / n)
		return NULL;
	return kzalloc(n * size, flags);
}

static inline void kfree (const void *pobj)
{
	ASSERT( KeGetCurrentIrql() <= DISPATCH_LEVEL );
	if (pobj)
		ExFreePool((void *)pobj);
}

#define get_zeroed_page(mask)			kzalloc(PAGE_SIZE, mask)
#define free_page(ptr) 					kfree(ptr)


////////////////////////////////////////////////////////
//
// IO SPACE <==> SYSTEM MEMORY
//
////////////////////////////////////////////////////////

/**
* ioremap     -   map bus memory into CPU space
* @addr:      bus address of the memory
* @size:      size of the resource to map
*
* ioremap performs a platform specific sequence of operations to
* make bus memory CPU accessible via the readb/readw/readl/writeb/
* writew/writel functions and the other mmio helpers. The returned
* address is not guaranteed to be usable directly as a virtual
* address. 
*/
static inline void *ioremap(io_addr_t addr, SIZE_T size, MEMORY_CACHING_TYPE cache_type)
{
	PHYSICAL_ADDRESS pa;
	void *va;
	
	ASSERT( KeGetCurrentIrql() <= DISPATCH_LEVEL );
	pa.QuadPart = addr;
	va = MmMapIoSpace( pa, size, cache_type ); 
	return va;
}

static inline void iounmap(void *va, SIZE_T size)
{
	MmUnmapIoSpace( va, size );
}


////////////////////////////////////////////////////////
//
// DMA SUPPORT
//
////////////////////////////////////////////////////////

enum dma_data_direction {
	PCI_DMA_BIDIRECTIONAL,
	PCI_DMA_TODEVICE,
	DMA_TO_DEVICE = PCI_DMA_TODEVICE
};

#define dma_get_cache_alignment		(int)KeGetRecommendedSharedDataAlignment

// wrapper to DMA address
typedef struct _dma_addr
{
	// TODO: in some cases it is still physical address today
	io_addr_t			da;		/* logical (device) address */
	void *				va;		/* kernel virtual address */
	unsigned long		sz; 	/* buffer size */
} dma_addr_t;

#define lowmem_page_address(dma_addr)	((dma_addr).va)

struct mlx4_dev;

void *alloc_cont_mem(
	IN		struct pci_dev *pdev, 
	IN		unsigned long size,
	OUT		dma_addr_t*p_dma_addr);

void free_cont_mem(
	IN		struct pci_dev *pdev, 
	IN		dma_addr_t*p_dma_addr);

// TODO: translate to DMA space - for now is not done anything
static inline dma_addr_t pci_map_page(struct pci_dev *pdev, 
	dma_addr_t dma_addr, unsigned long offset, SIZE_T size, int direction)
{
	UNUSED_PARAM(pdev);
	UNUSED_PARAM(offset);
	UNUSED_PARAM(size);
	UNUSED_PARAM(direction);

	return dma_addr; 
}

static inline dma_addr_t 
alloc_pages( struct pci_dev *pdev, gfp_t gfp, int order )
{
	dma_addr_t dma_addr;
	UNUSED_PARAM(gfp);
	alloc_cont_mem( pdev, PAGE_SIZE << order, &dma_addr );
	return dma_addr;
}

#define alloc_page(pdev, mask)		alloc_pages(pdev, (mask), 0)
#define __get_free_page(mask)		kzalloc(PAGE_SIZE, mask)

static inline void 
__free_pages( struct pci_dev *pdev, dma_addr_t dma_addr, int order )
{
	UNUSED_PARAM(order);
	ASSERT((PAGE_SIZE << order) == (int)dma_addr.sz);
	free_cont_mem( pdev, &dma_addr );
}

#define __free_page(pdev, dma_addr)	__free_pages(pdev, (dma_addr), 0)



static inline int pci_dma_mapping_error(dma_addr_t dma_addr)
{
	return !dma_addr.sz;
}

static inline void pci_unmap_page(struct pci_dev *pdev, 
	dma_addr_t dma_addr, SIZE_T size, int direction)
{
	UNUSED_PARAM(pdev);
	UNUSED_PARAM(dma_addr);
	UNUSED_PARAM(size);
	UNUSED_PARAM(direction);
}

static inline void
dma_sync_single( struct mlx4_dev **dev, dma_addr_t dma_addr,
	size_t size, int direction)
{
	UNUSED_PARAM(dev);
	UNUSED_PARAM(dma_addr);
	UNUSED_PARAM(size);
	UNUSED_PARAM(direction);
	// TODO: here is to be used FlushAdapterBuffers()
}

void *
dma_alloc_coherent( struct mlx4_dev **dev, size_t size, 
	dma_addr_t *p_dma, gfp_t gfp );

void dma_free_coherent( struct mlx4_dev **dev, size_t size, 
	void *vaddr, dma_addr_t dma_handle);

 void pci_free_consistent( struct pci_dev *pdev, size_t size, 
	 void *vaddr, dma_addr_t dma_handle);



////////////////////////////////////////////////////////
//
// SG lists
//
////////////////////////////////////////////////////////

#define sg_dma_addr(sg)					((sg)->dma_addr)
#define sg_dma_address(sg)				((sg)->dma_addr.da)
#define sg_dma_len(sg)					((sg)->dma_addr.sz)
#define sg_dma_address_inc(p_dma,val)	(p_dma)->da += val
#define sg_page(sg)						((sg)->dma_addr)

struct scatterlist {
	dma_addr_t 			dma_addr;	/* logical (device) address */
	unsigned int		offset;		/* offset in the first page */
	PMDL				p_mdl;		/* MDL, if any (used for user space buffers) */
};

#define offset_in_page(va)	((ULONG)((ULONG_PTR)(va) & ~PAGE_MASK))

static inline void sg_init_table(struct scatterlist *sgl, unsigned int nents)
{
	memset(sgl, 0, sizeof(*sgl) * nents);
}

static inline void sg_set_buf(struct scatterlist *sg, const void *buf,
	unsigned int buflen)
{
	UNUSED_PARAM(buflen);
	ASSERT(sg->dma_addr.sz == buflen);
	sg->offset = offset_in_page(buf);
}

static inline void sg_set_page(struct scatterlist *sg, 
	dma_addr_t dma_addr, unsigned int len, unsigned int offset)
{
	UNUSED_PARAM(len);
	sg->offset = offset;
	sg->dma_addr = dma_addr;
}

/* Returns: the number of unmapped sg elements */
static inline int pci_map_sg(struct pci_dev *pdev, 
	struct scatterlist *sg, int nents, int direction)
{
	UNUSED_PARAM(pdev);
	UNUSED_PARAM(sg);
	UNUSED_PARAM(direction);
	return nents;
}

/* Returns: the number of unmapped sg elements */
static inline int pci_unmap_sg(struct pci_dev *pdev, 
	struct scatterlist *sg, 	int nents, int direction)
{
	UNUSED_PARAM(pdev);
	UNUSED_PARAM(sg);
	UNUSED_PARAM(direction);
	return nents;
}

