/*
 * Copyright (c) 2005 Topspin Communications.  All rights reserved.
 * Copyright (c) 2006 Mellanox Technologies Ltd.  All rights reserved.
 * Copyright (c) 2008-2009 Intel Corporation.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AWV
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>

#include <string.h>
#include <malloc.h>
#include <getopt.h>
#include <time.h>
#include <infiniband/verbs.h>
#include <windows.h>

#include <ws2tcpip.h>
#include <winsock2.h>


#include "..\..\tools\perftests\user\get_clock.h"
#include "..\..\etc\user\getopt.c"
#include "perftest_resources.h"
#include "multicast_resources.h"
#include "l2w.h"

#define VERSION 1.3
#define SIGNAL 1
#define MAX_INLINE 400

static int page_size;
cycles_t  *tstamp;

struct report_options {
	int unsorted;
	int histogram;
	int cycles;   /* report delta's in cycles, not microsec's */
};

struct pingpong_context {
	struct ibv_sge      list;
	struct ibv_sge      recv_list;
	struct ibv_send_wr  wr;
	struct ibv_recv_wr  rwr;
	struct ibv_context  *context;
	struct ibv_comp_channel *channel;
	struct ibv_pd       *pd;
	struct ibv_mr       *mr;
	struct ibv_cq       *rcq;
	struct ibv_cq       *scq;
	struct ibv_qp       **qp;
	struct ibv_ah	    *ah;
	void                *buf;
	int                 size;	
};

/****************************************************************************** 
 *
 ******************************************************************************/
static int set_mcast_group(struct pingpong_context *ctx,
						   struct perftest_parameters *user_parm,
						   struct mcast_parameters *mcg_params) {

	int i;
	struct ibv_port_attr port_attr;

	if (ibv_query_gid(ctx->context,user_parm->ib_port,user_parm->gid_index,&mcg_params->port_gid)) {
			return 1;
	}
		
	if (ibv_query_pkey(ctx->context,user_parm->ib_port,DEF_PKEY_IDX,&mcg_params->pkey)) {
		return 1;
	}

	if (ibv_query_port(ctx->context,user_parm->ib_port,&port_attr)) {
		return 1;
	}
	mcg_params->sm_lid  = port_attr.sm_lid;
	mcg_params->sm_sl   = port_attr.sm_sl;
	mcg_params->ib_port = user_parm->ib_port;
	set_multicast_gid(mcg_params);

	// Request for Mcast group create registery in SM.
	if (join_multicast_group(SUBN_ADM_METHOD_SET,mcg_params)) {
		fprintf(stderr,"Bla Bla ....\n");
		return 1;
	}

	for (i=0; i < user_parm->num_of_qps; i++) {

		if (ibv_attach_mcast(ctx->qp[i],&mcg_params->mgid,mcg_params->mlid)) {
			fprintf(stderr, "Couldn't attach QP to MultiCast group");
			return 1;
		}
	}
	mcg_params->mcast_state |= MCAST_IS_ATTACHED;

	return 0;
}

/****************************************************************************** 
 *
 ******************************************************************************/
static int set_up_connection(struct pingpong_context *ctx,
							 struct perftest_parameters *user_parm,
							 struct pingpong_dest *my_dest,
							 struct mcast_parameters *mcg_params) {

	

	if (user_parm->use_mcg) {

		if (set_mcast_group(ctx,user_parm,mcg_params)) {
			return 1;
		}

		my_dest->gid = mcg_params->mgid;
		my_dest->lid = mcg_params->mlid;
		my_dest->qpn = QPNUM_MCAST;
	}
	else {
		if (user_parm->gid_index != -1) {
			if (ibv_query_gid(ctx->context,user_parm->ib_port,user_parm->gid_index,&my_dest->gid)) {
				return -1;
			}
		}
		my_dest->lid   	   = ctx_get_local_lid(ctx->context,user_parm->ib_port);
		my_dest->qpn   	   = ctx->qp[0]->qp_num;
	}

	my_dest->psn       = rand() & 0xffffff;
	my_dest->rkey      = ctx->mr->rkey;
	my_dest->vaddr     = (uintptr_t)ctx->buf + ctx->size;

	// We do not fail test upon lid above RoCE.
	if (user_parm->gid_index < 0) {
		if (!my_dest->lid) {
			fprintf(stderr,"Local lid 0x0 detected,without any use of gid. Is SM running?\n");
			return -1;
		}
	}
	return 0;
}

/****************************************************************************** 
 *
 ******************************************************************************/
static int init_connection(struct perftest_parameters *params,
 						   struct pingpong_dest *my_dest,
						   const char *servername) {

	params->side = LOCAL;

	ctx_print_pingpong_data(my_dest,params);
	
	if (params->machine == CLIENT) 
		params->sockfd = ctx_client_connect(servername,params->port);
	else 
		params->sockfd = ctx_server_connect(params->port);

	if (params->sockfd == INVALID_SOCKET) {
		fprintf(stderr,"Unable to open file descriptor for socket connection");
		return 1;
	}
	return 0;
}

/****************************************************************************** 
 *
 ******************************************************************************/
static struct ibv_device *pp_find_dev(const char *ib_devname) {
	struct ibv_device **dev_list;
	struct ibv_device *ib_dev = NULL;

	dev_list = ibv_get_device_list(NULL);

	if (!ib_devname) {
		ib_dev = dev_list[0];
		if (!ib_dev)
			fprintf(stderr, "No IB devices found\n");
	} else {
		for (; (ib_dev = *dev_list); ++dev_list)
			if (!strcmp(ibv_get_device_name(ib_dev), ib_devname))
				break;
		if (!ib_dev)
			fprintf(stderr, "IB device %s not found\n", ib_devname);
	}
	return ib_dev;
}

/****************************************************************************** 
 *
 ******************************************************************************/
static struct pingpong_context *pp_init_ctx(struct ibv_device *ib_dev, int size,
											struct perftest_parameters *user_parm) {

	struct pingpong_context *ctx;
	struct ibv_device_attr device_attr;
	int i;

	ALLOCATE(ctx,struct pingpong_context,1);

	ctx->ah 	  = NULL;
	ctx->size     = size;
 
	// Allocating the Buff size according to connection type and size.
	posix_memalign(&(ctx->buf),page_size, 2*SIZE(user_parm->connection_type,size));
	if (!ctx->buf) {
		fprintf(stderr, "Couldn't allocate work buf.\n");
		return NULL;
	}
	memset(ctx->buf, 0, 2*SIZE(user_parm->connection_type,size));

	ctx->context = ibv_open_device(ib_dev);
	if (!ctx->context) {
		fprintf(stderr, "Couldn't get context for %s\n",
			ibv_get_device_name(ib_dev));
		return NULL;
	}

	// Finds the link type and configure the HCA accordingly.
	if (ctx_set_link_layer(ctx->context,user_parm)) {
		fprintf(stderr, "Couldn't set the link layer\n");
		return NULL;
	}

	if (user_parm->mtu == 0) {/*user did not ask for specific mtu */
		if (ibv_query_device(ctx->context, &device_attr)) {
			fprintf(stderr, "Failed to query device props");
			return NULL;
		}
		if (device_attr.vendor_part_id == 23108 || user_parm->gid_index > -1) {
			user_parm->mtu = 1024;
		} else {
			user_parm->mtu = 2048;
		}
	}

    if (user_parm->use_event) {
		ctx->channel = ibv_create_comp_channel(ctx->context);
		if (!ctx->channel) {
			fprintf(stderr, "Couldn't create completion channel\n");
			return NULL;
		}
	} else
		ctx->channel = NULL;

	ctx->pd = ibv_alloc_pd(ctx->context);
	if (!ctx->pd) {
		fprintf(stderr, "Couldn't allocate PD\n");
		return NULL;
	}

	ctx->mr = ibv_reg_mr(ctx->pd,ctx->buf,2*SIZE(user_parm->connection_type,size),
						 IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_LOCAL_WRITE);
	if (!ctx->mr) {
		fprintf(stderr, "Couldn't allocate MR\n");
		return NULL;
	}
	
	ctx->scq = ibv_create_cq(ctx->context,user_parm->tx_depth,NULL,ctx->channel,0);
	if (!ctx->scq) {
	    fprintf(stderr, "Couldn't create CQ\n");
		return NULL;
	}

	ctx->rcq = ibv_create_cq(ctx->context,user_parm->rx_depth,NULL,ctx->channel,0);
	if (!ctx->rcq) {
	    fprintf(stderr, "Couldn't create CQ\n");
		return NULL;
	}

	ALLOCATE(ctx->qp,struct ibv_qp*,user_parm->num_of_qps);

	for (i = 0; i < user_parm->num_of_qps; i++) {
		
		ctx->qp[i] = ctx_qp_create(ctx->pd,ctx->scq,ctx->rcq,user_parm);
		if (ctx->qp[i] == NULL) {
			return NULL;
		}

		if(ctx_modify_qp_to_init(ctx->qp[i],user_parm)) {
			return NULL;
		}
	}
	return ctx;
}

/****************************************************************************** 
 *
 ******************************************************************************/
static int pp_connect_ctx(struct pingpong_context *ctx,int my_psn,
						  struct pingpong_dest *dest,
						  struct perftest_parameters *user_parm)
{
	struct ibv_qp_attr attr;
	struct ibv_recv_wr *bad_wr_recv;
	int i,j;

	memset(&attr, 0, sizeof(struct ibv_qp_attr));
	attr.qp_state                 = IBV_QPS_RTR;
	switch (user_parm->mtu) {
		case 256 : 
			attr.path_mtu               = IBV_MTU_256;
			break;
		case 512 :
			attr.path_mtu               = IBV_MTU_512;
			break;
		case 1024 :
			attr.path_mtu               = IBV_MTU_1024;
			break;
		case 2048 :
			attr.path_mtu               = IBV_MTU_2048;
			break;
		case 4096 :
			attr.path_mtu               = IBV_MTU_4096;
			break;
	}
	attr.dest_qp_num      = dest->qpn;
	attr.rq_psn           = dest->psn;
	attr.ah_attr.dlid     = dest->lid;
	printf("Mtu : %d\n", user_parm->mtu);
	if (user_parm->connection_type==RC) {
		attr.max_dest_rd_atomic     = 1;
		attr.min_rnr_timer          = 12;
	}

	if (user_parm->gid_index < 0) {
		attr.ah_attr.is_global      = 0;
		attr.ah_attr.sl             = (uint8_t)(user_parm->sl);
	} else {
		attr.ah_attr.is_global      = 1;
		attr.ah_attr.grh.dgid       = dest->gid;
		attr.ah_attr.grh.sgid_index = (uint8_t)(user_parm->gid_index);
		attr.ah_attr.grh.hop_limit  = 1;
		attr.ah_attr.sl             = 0;
	}
	attr.ah_attr.src_path_bits  = 0;
	attr.ah_attr.port_num       = user_parm->ib_port;

	if (user_parm->connection_type==RC) {
		if (ibv_modify_qp(ctx->qp[0], &attr,
				  IBV_QP_STATE              |
				  IBV_QP_AV                 |
				  IBV_QP_PATH_MTU           |
				  IBV_QP_DEST_QPN           |
				  IBV_QP_RQ_PSN             |
				  IBV_QP_MIN_RNR_TIMER      |
				  IBV_QP_MAX_DEST_RD_ATOMIC)) {
			fprintf(stderr, "Failed to modify RC QP to RTR\n");
			return 1;
		}
		attr.timeout            = user_parm->qp_timeout;
		attr.retry_cnt          = 7;
		attr.rnr_retry          = 7;
	} else if (user_parm->connection_type==UC) {
		if (ibv_modify_qp(ctx->qp[0], &attr,
				  IBV_QP_STATE              |
				  IBV_QP_AV                 |
				  IBV_QP_PATH_MTU           |
				  IBV_QP_DEST_QPN           |
				  IBV_QP_RQ_PSN)) {
				fprintf(stderr, "Failed to modify UC QP to RTR\n");
				return 1;
		}

	} else {
		for (i = 0; i < user_parm->num_of_qps; i++) {
			if (ibv_modify_qp(ctx->qp[i],&attr,IBV_QP_STATE )) {
				fprintf(stderr, "Failed to modify UC QP to RTR\n");
				return 1;
			}
		}

		ctx->ah = ibv_create_ah(ctx->pd,&attr.ah_attr);
		if (!ctx->ah) {
			fprintf(stderr, "Failed to create AH for UD\n");
            return 1;
		}
	}

	attr.qp_state             = IBV_QPS_RTS;
	attr.sq_psn      		  = my_psn;
	if (user_parm->connection_type==RC) {
		attr.max_rd_atomic  = 1;
		if (ibv_modify_qp(ctx->qp[0], &attr,
				  IBV_QP_STATE              |
				  IBV_QP_SQ_PSN             |
				  IBV_QP_TIMEOUT            |
				  IBV_QP_RETRY_CNT          |
				  IBV_QP_RNR_RETRY          |
				  IBV_QP_MAX_QP_RD_ATOMIC)) {
			fprintf(stderr, "Failed to modify RC QP to RTS\n");
			return 1;
		}
	} else { 

		if(ibv_modify_qp(ctx->qp[0],&attr,IBV_QP_STATE |IBV_QP_SQ_PSN)) {
			fprintf(stderr, "Failed to modify UC QP to RTS\n");
			return 1;
		}
	}

	// Receive 
	ctx->rwr.sg_list      = &ctx->recv_list;
	ctx->rwr.num_sge      = MAX_RECV_SGE;
	ctx->rwr.next         = NULL;
	ctx->recv_list.lkey   = ctx->mr->lkey;
	ctx->recv_list.addr   = (uintptr_t) ctx->buf;
	ctx->recv_list.length = SIZE(user_parm->connection_type,ctx->size);

	// Send
	ctx->wr.sg_list    = &ctx->list;
	ctx->wr.num_sge    = MAX_SEND_SGE;
	ctx->wr.next       = NULL;
	ctx->wr.opcode     = IBV_WR_SEND;
	ctx->wr.wr_id      = PINGPONG_SEND_WRID;
	ctx->list.lkey     = ctx->mr->lkey;
	ctx->list.addr 	   = (uintptr_t)ctx->buf + SIZE(user_parm->connection_type,ctx->size);

	if (user_parm->connection_type == UD) {
		ctx->wr.wr.ud.ah          = ctx->ah;
		ctx->wr.wr.ud.remote_qpn  = dest->qpn;
		ctx->wr.wr.ud.remote_qkey = DEF_QKEY;
	}

	// Post Receiving rx-depth to keep the Queue full.
	for (i = 0; i < user_parm->num_of_qps; i++) {
		ctx->rwr.wr_id = i;
		for (j = 0; j < (user_parm->rx_depth / 2); j++ ) {
			if (ibv_post_recv(ctx->qp[i],&ctx->rwr, &bad_wr_recv)) {
				fprintf(stderr, "Couldn't post recv: counter=%d\n", j);
				return 14;
			}
		}
	}
	
	return 0;
}

static void usage(const char *argv0)
{
	printf("Usage:\n");
	printf("  %s            start a server and wait for connection\n", argv0);
	printf("  %s <host>     connect to server at <host>\n", argv0);
	printf("\n");
	printf("Options:\n");
	printf("  -p, --port=<port>            Listen on/connect to port <port> (default 18515)\n");
	printf("  -c, --connection=<RC/UC/UD>  Connection type RC/UC/UD (default RC)\n");
	printf("  -m, --mtu=<mtu>              Mtu size (256 - 4096. default for hermon is 2048)\n");
	printf("  -d, --ib-dev=<dev>           Use IB device <dev> (default first device found)\n");
	printf("  -i, --ib-port=<port>         Use port <port> of IB device (default 1)\n");
	printf("  -s, --size=<size>            Size of message to exchange (default 1)\n");
	printf("  -t, --tx-depth=<dep>         Size of tx queue (default 50)\n");
	printf("  -l, --signal                 Signal completion on each msg\n");
	printf("  -a, --all                    Run sizes from 2 till 2^23\n");
	printf("  -n, --iters=<iters>          Number of exchanges (at least 2, default 1000)\n");
	printf("  -I, --inline_size=<size>     Max size of message to be sent in inline mode (default 400)\n");
	printf("  -u, --qp-timeout=<timeout>   QP timeout, timeout value is 4 usec * 2 ^(timeout), default 14\n");
	printf("  -S, --sl=<sl>                SL (default 0)\n");
	printf("  -x, --gid-index=<index>      Test uses GID with GID index taken from command line (for RDMAoE index should be 0)\n");
	printf("  -C, --report-cycles          Report times in cpu cycle units (default microseconds)\n");
	printf("  -H, --report-histogram       Print out all results (default print summary only)\n");
	printf("  -U, --report-unsorted        (implies -H) print out unsorted results (default sorted)\n");
	printf("  -V, --version                Display version number\n");
	printf("  -e, --events                 Sleep on CQ events (default poll)\n");
	printf("  -g, --mcg=<num_of_qps>       Send messages to multicast group with <num_of_qps> qps attached to it.\n");
	printf("  -M, --MGID=<multicast_gid>   In case of multicast, uses <multicast_gid> as the group MGID.\n");
	printf("                               The format must be '255:1:X:X:X:X:X:X:X:X:X:X:X:X:X:X', where X is a vlaue within [0,255].\n");
	printf("                               You must specify a different MGID on both sides (to avoid loopback).\n");
	printf("  -F, --CPU-freq               Do not fail even if cpufreq_ondemand module is loaded\n");
}

static void print_report(struct report_options *options,
			 unsigned int iters, cycles_t *tstamp,int size, int no_cpu_freq_fail)
{
	double cycles_to_units;
	cycles_t median;
	unsigned int i;
	const char* units;
	cycles_t *delta = malloc((iters - 1) * sizeof *delta);

	if (!delta) {
		perror("malloc");
		return;
	}

	for (i = 0; i < iters - 1; ++i)
		delta[i] = tstamp[i + 1] - tstamp[i];


	if (options->cycles) {
		cycles_to_units = 1;
		units = "cycles";
	} else {
		cycles_to_units = get_cpu_mhz()/1000000;
		units = "usec";
	}

	if (options->unsorted) {
		printf("#, %s\n", units);
		for (i = 0; i < iters - 1; ++i)
			printf("%d, %g\n", i + 1, delta[i] / cycles_to_units / 2);
	}

	qsort(delta, iters - 1, sizeof *delta, cycles_compare);

	if (options->histogram) {
		printf("#, %s\n", units);
		for (i = 0; i < iters - 1; ++i)
			printf("%d, %g\n", i + 1, delta[i] / cycles_to_units / 2);
	}

	median = get_median(iters - 1, delta);
	printf("%7d        %d        %7.2f        %7.2f          %7.2f\n",
	       size,iters,delta[0] / cycles_to_units / 2,
	       delta[iters - 2] / cycles_to_units / 2,median / cycles_to_units / 2);
	free(delta);
}

/* 
 *
 */
int run_iter(struct pingpong_context *ctx, struct perftest_parameters *user_param,
			 struct pingpong_dest *rem_dest, int size)
{
	
	struct ibv_wc 			*wc;
	struct ibv_recv_wr      *bad_wr_recv;
	struct ibv_send_wr 		*bad_wr;
	int                     i,scnt,rcnt,poll,qp_counter,ne;

	ALLOCATE(wc,struct ibv_wc,user_param->num_of_qps);

	
	ctx->list.length   = size;
	ctx->wr.send_flags = size > user_param->inline_size ? 0 : IBV_SEND_INLINE;

	scnt = 0;
	rcnt = 0;
	poll = 0;
	qp_counter = 0;
	
	while (scnt < user_param->iters || rcnt < user_param->iters) {

		if (rcnt < user_param->iters && !(scnt < 1 && user_param->machine == CLIENT)) {
		  
			// Server is polling on recieve first .
		    if (user_param->use_event) {
				struct ibv_cq *ev_cq;
				void          *ev_ctx;

				if (ibv_get_cq_event(ctx->channel, &ev_cq, &ev_ctx)) {
					fprintf(stderr, "Failed to get receive cq_event\n");
					return 1;
				}
			    
				if (ev_cq != ctx->rcq) {
					fprintf(stderr, "CQ event for unknown RCQ %p\n", ev_cq);
					return 1;
				}

				if (ibv_req_notify_cq(ctx->rcq, 0)) {
					fprintf(stderr, "Couldn't request RCQ notification\n");
					return 1;
				}
		    }

			do {
				ne = ibv_poll_cq(ctx->rcq, user_param->num_of_qps, wc);
				if (ne > 0) {
					for (i = 0; i < ne; i++) {

						if (wc[i].status != IBV_WC_SUCCESS) {
							fprintf(stderr, "Completion wth error at %s:\n",user_param->machine == CLIENT ? "client" : "server");
							fprintf(stderr, "Failed status %d: wr_id %d syndrom 0x%x\n",wc[i].status, (int)wc[i].wr_id, wc[i].vendor_err);
							fprintf(stderr, "rcnt=%d\n",rcnt);
							return 1;
						}

						qp_counter++;
						ctx->rwr.wr_id = wc[i].wr_id;
						if (ibv_post_recv(ctx->qp[wc[i].wr_id], &ctx->rwr, &bad_wr_recv)) {
							fprintf(stderr, "Couldn't post recv: rcnt=%d\n",rcnt);
							return 15;
						}
					}
				}
			} while (!user_param->use_event && qp_counter < user_param->num_of_qps);
			rcnt++;
			// printf(" Got completion %d \n",rcnt);
			qp_counter  = 0;
		}

		// client post first. 
		if (scnt < user_param->iters) {
			tstamp[scnt++] = get_cycles();
			if (scnt  == user_param->iters || scnt%user_param->rx_depth == 0) {
				poll = 1;
				ctx->wr.send_flags |= IBV_SEND_SIGNALED;
			}
			
			if (ibv_post_send(ctx->qp[0],&ctx->wr,&bad_wr)) {
				fprintf(stderr, "Couldn't post send: scnt=%d\n",scnt);
				return 11;
			}
		}

		if (poll == 1) {

		    struct ibv_wc s_wc;
		    int s_ne;

		    if (user_param->use_event) {
				struct ibv_cq *ev_cq;
				void          *ev_ctx;

				if (ibv_get_cq_event(ctx->channel, &ev_cq, &ev_ctx)) {
					fprintf(stderr, "Failed to get send cq_event\n");
					return 1;
				}

				if (ev_cq != ctx->scq) {
					fprintf(stderr, "CQ event for unknown SCQ %p\n", ev_cq);
					return 1;
				}

				if (ibv_req_notify_cq(ctx->scq, 0)) {
					fprintf(stderr, "Couldn't request SCQ notification\n");
					return 1;
				}
		    }
		    
		    do {
				s_ne = ibv_poll_cq(ctx->scq, 1, &s_wc);
		    } while (!user_param->use_event && s_ne == 0);

		    if (s_ne < 0) {
				fprintf(stderr, "poll SCQ failed %d\n", s_ne);
				return 12;
		    }

			if (s_wc.status != IBV_WC_SUCCESS) {
				fprintf(stderr, "Completion wth error at %s:\n",user_param->machine == CLIENT ? "client" : "server");
				fprintf(stderr, "Failed status %d: wr_id %d\n",s_wc.status, (int) s_wc.wr_id);
				fprintf(stderr, "scnt=%d, rcnt=%d,\n" ,scnt, rcnt);
				return 13;
		    }

			poll = 0;
			ctx->wr.send_flags &= ~IBV_SEND_SIGNALED;
		}
	}
	return 0;
}

int __cdecl main(int argc, char *argv[])
{
	const char                 *ib_devname = NULL;
	int                        size = 2;
	int                        i = 0;
	int                        size_max_pow = 24;
	struct report_options      report;
	struct pingpong_context    *ctx = NULL;
	struct pingpong_dest	   my_dest,rem_dest;
	struct mcast_parameters	   mcg_params;
	struct ibv_device          *ib_dev = NULL;
	struct perftest_parameters user_param;
	int                        no_cpu_freq_fail = 0;

	int all = 0;
	const char *servername = NULL;
	
	SYSTEM_INFO si;
	GetSystemInfo(&si);

	/* init default values to user's parameters */
	memset(&user_param, 0, sizeof(struct perftest_parameters));
	memset(&mcg_params, 0, sizeof(struct mcast_parameters));
	memset(&my_dest   , 0 , sizeof(struct pingpong_dest));
	memset(&rem_dest   , 0 , sizeof(struct pingpong_dest));

	user_param.iters = 1000;
	user_param.port = 18515;
	user_param.ib_port = 1;
	user_param.tx_depth = 50;
	user_param.rx_depth = 50;
	user_param.inline_size = MAX_INLINE;
	user_param.verb = SEND;
	user_param.qp_timeout = 14;
	user_param.gid_index = -1; /*gid will not be used*/
	user_param.num_of_qps = 1;

	/* Parameter parsing. */
	while (1) {
		int c;

		static struct option long_options[] = {
			{ "port",             1, NULL, 'p' },
			{ "connection",       1, NULL, 'c' },
			{ "mtu",              1, NULL, 'm' },
			{ "ib-dev",           1, NULL, 'd' },
			{ "ib-port",          1, NULL, 'i' },
			{ "size",             1, NULL, 's' },
			{ "iters",            1, NULL, 'n' },
			{ "tx-depth",         1, NULL, 't' },
			{ "inline_size",      1, NULL, 'I' },
			{ "qp-timeout",       1, NULL, 'u' },
			{ "sl",               1, NULL, 'S' },
			{ "gid-index",        1, NULL, 'x' },
			{ "signal",           0, NULL, 'l' },
			{ "all",              0, NULL, 'a' },
			{ "report-cycles",    0, NULL, 'C' },
			{ "report-histogram", 0, NULL, 'H'},
			{ "report-unsorted",  0, NULL, 'U' },
			{ "version",          0, NULL, 'V' },
			{ "events",           0, NULL, 'e' },
			{ "mcg",              1, NULL, 'g' },
			{ "MGID",             1, NULL, 'M' },
			{ "CPU-freq",         0, NULL, 'F' },
			{ 0 }
		};
		c = getopt_long(argc, argv, "p:c:m:d:i:s:n:t:I:u:S:x:g:M:laeCHUVF", long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'p':
			user_param.port = strtol(optarg, NULL, 0);
			if (user_param.port < 0 || user_param.port > 65535) {
				usage(argv[0]);
				return 1;
			}
			break;
		case 'c':
			if (strcmp("UC",optarg)==0)
				user_param.connection_type=UC;
			if (strcmp("UD",optarg)==0)
				user_param.connection_type=UD;
			/* default is 0 for any other option RC*/
			break;
        case 'e':
			++user_param.use_event;
			break;
		case 'g':
			++user_param.use_mcg;
			user_param.num_of_qps = strtol(optarg, NULL, 0);
			if (user_param.num_of_qps < 1 || user_param.num_of_qps > 57) {
				usage(argv[0]);
				return 1;
			}
			break;
		case 'M' :
			mcg_params.is_user_mgid = 1;
			mcg_params.user_mgid = _strdup(optarg);
			break;
		case 'm':
			user_param.mtu = strtol(optarg, NULL, 0);
			break;
		case 'l':
			user_param.signal_comp = SIGNAL;
			break;
		case 'a':
			all = ALL;
			break;
		case 'V':
			printf("perftest version : %.2f\n",VERSION);
			return 0;
			break;
		case 'd':
			ib_devname = _strdup(optarg);
			break;

		case 'i':
			user_param.ib_port = (uint8_t)(strtol(optarg, NULL, 0));
			if (user_param.ib_port < 0) {
				usage(argv[0]);
				return 2;
			}
			break;

		case 's':
			size = strtol(optarg, NULL, 0);
			if (size < 1) {
				usage(argv[0]); return 3;
			}
			break;

		case 'x':
			user_param.gid_index = strtol(optarg, NULL, 0);
			if (user_param.gid_index > 63) {
				usage(argv[0]);
				return 1;
			}
			break;

		case 't':
			user_param.tx_depth = strtol(optarg, NULL, 0);
			user_param.rx_depth = user_param.tx_depth;
			if (user_param.tx_depth < 1) {
				usage(argv[0]); return 4;
			}
			break;

		case 'I':
			user_param.inline_size = strtol(optarg, NULL, 0);
			if (user_param.inline_size > MAX_INLINE) {
				usage(argv[0]);
				return 19;
			}
			break;

		case 'n':
			user_param.iters = strtol(optarg, NULL, 0);
			if (user_param.iters < 2) {
				usage(argv[0]);
				return 5;
			}

			break;

		case 'C':
			report.cycles = 1;
			break;

		case 'H':
			report.histogram = 1;
			break;

		case 'U':
			report.unsorted = 1;
			break;

		case 'F':
			no_cpu_freq_fail = 1;
			break;

		case 'u':
			user_param.qp_timeout = (uint8_t)(strtol(optarg, NULL, 0));
			break;

		case 'S':
			user_param.sl = strtol(optarg, NULL, 0);
			if (user_param.sl > 15) { usage(argv[0]); return 6; }
			break;

		default:
			usage(argv[0]);
			return 7;
		}
	}

	if (optind == argc - 1)
		servername = _strdup(argv[optind]);
	else if (optind < argc) {
		usage(argv[0]);
		return 6;
	}

	user_param.machine = servername ? CLIENT : SERVER;

	/*
	 *  Done with parameter parsing. Perform setup.
	 */
	tstamp = malloc(user_param.iters * sizeof *tstamp);
	if (!tstamp) {
		perror("malloc");
		return 10;
	}
	/* Print header data */
	printf("------------------------------------------------------------------\n");
	if (user_param.use_mcg) { 
		user_param.connection_type = UD;
		printf("                    Send Latency Multicast Test\n");
	} else {
		printf("                    Send Latency Test\n");
	}

	printf("Inline data is used up to %d bytes message\n", user_param.inline_size);
	if (user_param.connection_type==RC) {
		printf("Connection type : RC\n");
	} else if (user_param.connection_type==UC) { 
		printf("Connection type : UC\n");
	} else {
		printf("Connection type : UD\n");
	}

	if (all == ALL) {
		/*since we run all sizes lets allocate big enough buffer */
		size = 8388608; /*2^23 */
	}
	if (user_param.connection_type == UD && size > 2048) {
		printf("Max msg size in UD is 2048 changing to 2048\n");
		size = 2048;
	}
	if (user_param.connection_type == UD && user_param.gid_index > -1 && size > 1024) {
		printf("Max msg size in UD RDMAoE is 1024. changing to 1024\n");
		size = 1024;
	}

	page_size = si.dwPageSize;

	ib_dev = pp_find_dev(ib_devname);
	if (!ib_dev)
		return 7;

	mcg_params.ib_devname = ibv_get_device_name(ib_dev);

	ctx = pp_init_ctx(ib_dev,size,&user_param);
	if (!ctx)
		return 8;

	// Set up the Connection.
	if (set_up_connection(ctx,&user_param,&my_dest,&mcg_params)) {
		fprintf(stderr," Unable to set up socket connection\n");
		return 1;
	}	

	// Init the connection and print the local data.
	if (init_connection(&user_param,&my_dest,servername)) {
		fprintf(stderr," Unable to init the socket connection\n");
		return 1;
	}

	// shaking hands and gather the other side info.
    if (ctx_hand_shake(&user_param,&my_dest,&rem_dest)) {
        fprintf(stderr,"Failed to exchange date between server and clients\n");
        return 1;
        
    }
	// For printing only MGID in the remote side.
	user_param.side = REMOTE;
	ctx_print_pingpong_data(&rem_dest,&user_param);

	// Connects......
	if (pp_connect_ctx(ctx,my_dest.psn,&rem_dest,&user_param)) {
		fprintf(stderr," Unable to Connect the HCA's through the link\n");
		return 1;
	}

	// An additional handshake is required after moving qp to RTR.
	if (ctx_hand_shake(&user_param,&my_dest,&rem_dest)) {
        fprintf(stderr,"Failed to exchange date between server and clients\n");
        return 1;
    }

    if (user_param.use_event) {
        printf("Test with events.\n");
        if (ibv_req_notify_cq(ctx->rcq, 0)) {
			fprintf(stderr, "Couldn't request RCQ notification\n");
			return 1;
		} 
        if (ibv_req_notify_cq(ctx->scq, 0)) {
			fprintf(stderr, "Couldn't request SCQ notification\n");
			return 1;
		}

    }
	printf("------------------------------------------------------------------\n");
	printf(" #bytes #iterations    t_min[usec]    t_max[usec]  t_typical[usec]\n");
    
	if (all == ALL) {
		if (user_param.connection_type==UD) {
			if (user_param.gid_index < 0) {
				size_max_pow = 12;
			} else {
				size_max_pow = 11;
			}
		}
		for (i = 1; i < size_max_pow ; ++i) {
			size = 1 << i;
			if(run_iter(ctx, &user_param, &rem_dest, size))
				return 17;

			print_report(&report,user_param.iters,tstamp,size,no_cpu_freq_fail);

			if (ctx_hand_shake(&user_param,&my_dest,&rem_dest)) {
				fprintf(stderr,"Failed to exchange date between server and clients\n");
				return 1;
			}
		}
	} else {
		if(run_iter(ctx,&user_param,&rem_dest,size))
			return 18;	
		print_report(&report, user_param.iters, tstamp, size, no_cpu_freq_fail);
	}

	if (ctx_close_connection(&user_param,&my_dest,&rem_dest)) {
		fprintf(stderr,"Failed to close connection between server and client\n");
		return 1;
	}

	printf("------------------------------------------------------------------\n");
	return 0;
}
