/*
 * Copyright (c) 2008 Microsoft Corporation.  All rights reserved.
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id:$
 */

#include <tchar.h>
#include <ndspi.h>
#include <iba/ib_at_ioctl.h>
#include <complib/cl_types.h>
#include <complib/cl_ioctl.h>
#pragma warning( push, 3 )
#include <unknwn.h>
#include <assert.h>
#include <ws2tcpip.h>
#include <winioctl.h>
#include <limits.h>
#include <ws2spi.h>
#pragma warning( pop )
#include "ndprov.h"
#include "ndadapter.h"

#if defined(EVENT_TRACING)
#ifdef offsetof
#undef offsetof
#endif
#include "NdProv.tmh"
#endif

#include "nddebug.h"

uint32_t g_nd_dbg_level = TRACE_LEVEL_ERROR;
/* WPP doesn't want here literals! */
uint32_t g_nd_dbg_flags = 0x80000001; /* ND_DBG_ERROR | ND_DBG_NDI; */
uint32_t g_nd_max_inline_size = 160;

HANDLE ghHeap;


namespace NetworkDirect
{

    static LONG gnRef = 0;

    CProvider::CProvider() :
        m_nRef( 1 )
    {
        InterlockedIncrement( &gnRef );
    }

    CProvider::~CProvider()
    {
        InterlockedDecrement( &gnRef );
    }

    HRESULT CProvider::QueryInterface(
        const IID &riid,
        void **ppObject )
    {
        if( IsEqualIID( riid, IID_IUnknown ) )
        {
            *ppObject = this;
            return S_OK;
        }

        if( IsEqualIID( riid, IID_INDProvider ) )
        {
            *ppObject = this;
            return S_OK;
        }

        return E_NOINTERFACE;
    }

    ULONG CProvider::AddRef()
    {
        return InterlockedIncrement( &m_nRef );
    }

    ULONG CProvider::Release()
    {
        ULONG ref = InterlockedDecrement( &m_nRef );
        if( ref == 0 )
            delete this;

        return ref;
    }

    HRESULT CProvider::QueryAddressList(
            __out_bcount_part_opt(*pBufferSize, *pBufferSize) SOCKET_ADDRESS_LIST* pAddressList,
            __inout SIZE_T* pBufferSize )
    {
        ND_ENTER( ND_DBG_NDI );

        HANDLE hIbatDev = CreateFileW( IBAT_WIN32_NAME,
            MAXIMUM_ALLOWED, 0, NULL,
            OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
        if( hIbatDev == INVALID_HANDLE_VALUE )
            return ND_NO_MEMORY;

        IOCTL_IBAT_IP_ADDRESSES_IN addrIn;

        addrIn.Version = IBAT_IOCTL_VERSION;
        addrIn.PortGuid = 0;

        DWORD size = sizeof(IOCTL_IBAT_IP_ADDRESSES_OUT);
        IOCTL_IBAT_IP_ADDRESSES_OUT *pAddrOut;
        do
        {
            pAddrOut = (IOCTL_IBAT_IP_ADDRESSES_OUT*)HeapAlloc(
                GetProcessHeap(),
                0,
                size );
            if( !pAddrOut )
            {
                //AL_PRINT( TRACE_LEVEL_ERROR, AL_DBG_ERROR,
                //    ("Failed to allocate output buffer.\n") );
                return ND_NO_MEMORY;
            }

            if( !DeviceIoControl( hIbatDev, IOCTL_IBAT_IP_ADDRESSES,
                &addrIn, sizeof(addrIn), pAddrOut, size, &size, NULL ) )
            {
                HeapFree( GetProcessHeap(), 0, pAddrOut );
                //AL_PRINT( TRACE_LEVEL_ERROR, AL_DBG_ERROR,
                //    ("IOCTL_IBAT_IP_ADDRESSES failed (%x).\n", GetLastError()) );
                return ND_UNSUCCESSFUL;
            }

            if( pAddrOut->Size > size )
            {
                size = pAddrOut->Size;
                HeapFree( GetProcessHeap(), 0, pAddrOut );
                pAddrOut = NULL;
            }

        } while( !pAddrOut );

        CloseHandle( hIbatDev );

        //
        // Note: the required size computed is a few bytes larger than necessary, 
        // but that keeps the code clean.
        //
        SIZE_T size_req = sizeof(SOCKET_ADDRESS_LIST);

        switch( pAddrOut->AddressCount )
        {
        case 0:
            break;

        default:
            size_req += (pAddrOut->AddressCount - 1) *
                (sizeof(SOCKET_ADDRESS) + sizeof(SOCKADDR));
            /* Fall through. */
            __fallthrough;

        case 1:
            /* Add the space for the first address. */
            size_req += sizeof(SOCKADDR);
            break;
        }

        if( size_req > *pBufferSize )
        {
            HeapFree( GetProcessHeap(), 0, pAddrOut );
            *pBufferSize = size_req;
            return ND_BUFFER_OVERFLOW;
        }

        ZeroMemory( pAddressList, *pBufferSize );

        /* We store the array of addresses after the last address pointer:
        *      iAddressCount
        *      Address[0]; <-- points to sockaddr[0]
        *      Address[1]; <-- points to sockaddr[1]
        *      ...
        *      Address[n-1]; <-- points to sockaddr[n-1]
        *      sockaddr[0];
        *      sockaddr[1];
        *      ...
        *      sockaddr[n-1]
        */
        BYTE* pBuf = (BYTE*)(&(pAddressList->Address[pAddrOut->AddressCount]));
        *pBufferSize = size_req;

        pAddressList->iAddressCount = 0;
        for( LONG i = 0; i < pAddrOut->AddressCount; i++ )
        {
            pAddressList->Address[pAddressList->iAddressCount].lpSockaddr =
                (LPSOCKADDR)pBuf;

            switch( pAddrOut->Address[i].IpVersion )
            {
            case 4:
                {
                    struct sockaddr_in* pAddr4 = ((struct sockaddr_in*)pBuf);
                    pAddr4->sin_family = AF_INET;
                    pAddr4->sin_addr.s_addr =
                        *((u_long*)&pAddrOut->Address[i].Address[12]);
                    pAddressList->Address[pAddressList->iAddressCount].iSockaddrLength =
                        sizeof(struct sockaddr_in);
                }
                break;

            case 6:
                {
                    struct sockaddr_in6* pAddr6 = ((struct sockaddr_in6*)pBuf);
                    pAddr6->sin6_family = AF_INET6;
                    CopyMemory(
                        &pAddr6->sin6_addr,
                        pAddrOut->Address[i].Address,
                        sizeof(pAddr6->sin6_addr) );
                    pAddressList->Address[pAddressList->iAddressCount].iSockaddrLength =
                        sizeof(struct sockaddr_in6);
                }
                break;

            default:
                continue;
            }

            pBuf += pAddressList->Address[pAddressList->iAddressCount++].iSockaddrLength;
        }

        HeapFree( GetProcessHeap(), 0, pAddrOut );

        return S_OK;
    }

    HRESULT CProvider::OpenAdapter(
            __in_bcount(AddressLength) const struct sockaddr* pAddress,
            __in SIZE_T AddressLength,
            __deref_out INDAdapter** ppAdapter )
    {
        ND_ENTER( ND_DBG_NDI );

        if( AddressLength < sizeof(struct sockaddr) )
            return ND_INVALID_ADDRESS;

        IOCTL_IBAT_IP_TO_PORT_IN in;
        in.Version = IBAT_IOCTL_VERSION;

        switch( pAddress->sa_family )
        {
        case AF_INET:
            if( AddressLength < sizeof(struct sockaddr_in) )
                return ND_INVALID_ADDRESS;
            in.Address.IpVersion = 4;
            RtlCopyMemory(
                &in.Address.Address[12],
                &((struct sockaddr_in*)pAddress)->sin_addr,
                sizeof( ((struct sockaddr_in*)pAddress)->sin_addr ) );
            break;

        case AF_INET6:
            if( AddressLength < sizeof(struct sockaddr_in6) )
                return ND_INVALID_ADDRESS;
            in.Address.IpVersion = 6;
            RtlCopyMemory(
                in.Address.Address,
                &((struct sockaddr_in6*)pAddress)->sin6_addr,
                sizeof(in.Address.Address) );
            break;

        default:
            return ND_INVALID_ADDRESS;
        }

        HANDLE hIbatDev = CreateFileW( IBAT_WIN32_NAME,
            MAXIMUM_ALLOWED, 0, NULL,
            OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
        if( hIbatDev == INVALID_HANDLE_VALUE )
            return ND_NO_MEMORY;

        IBAT_PORT_RECORD out;
        DWORD size;
        BOOL fSuccess = DeviceIoControl( hIbatDev, IOCTL_IBAT_IP_TO_PORT,
            &in, sizeof(in), &out, sizeof(out), &size, NULL );
        
        CloseHandle( hIbatDev );
        if( !fSuccess || size == 0 )
            return ND_INVALID_ADDRESS;

        return CAdapter::Create( this, pAddress, &out, ppAdapter );
    }

    CClassFactory::CClassFactory(void) :
        m_nRef( 1 )
    {
        InterlockedIncrement( &gnRef );
    }

    CClassFactory::~CClassFactory(void)
    {
        InterlockedDecrement( &gnRef );
    }

    HRESULT CClassFactory::QueryInterface(
        REFIID riid,
        void** ppObject )
    {
        if( IsEqualIID( riid, IID_IUnknown ) )
        {
            *ppObject = this;
            return S_OK;
        }
        if( IsEqualIID( riid, IID_IClassFactory ) )
        {
            *ppObject = this;
            return S_OK;
        }

        return E_NOINTERFACE;
    }

    ULONG CClassFactory::AddRef()
    {
        return InterlockedIncrement( &m_nRef );
    }

    ULONG CClassFactory::Release()
    {
        ULONG ref = InterlockedDecrement( &m_nRef );
        if( ref == 0 )
            delete this;

        return ref;
    }

    HRESULT CClassFactory::CreateInstance(
        IUnknown* pUnkOuter,
        REFIID riid,
        void** ppObject )
    {
        if( pUnkOuter != NULL )
            return CLASS_E_NOAGGREGATION;

        if( IsEqualIID( riid, IID_INDProvider ) )
        {
            *ppObject = new CProvider();
            if( !*ppObject )
                return E_OUTOFMEMORY;

            return S_OK;
        }

        return E_NOINTERFACE;
    }

    HRESULT CClassFactory::LockServer( BOOL fLock )
    { 
        UNREFERENCED_PARAMETER( fLock );
        return S_OK;
    }

} // namespace

void* __cdecl operator new(
    size_t count
    )
{
    return HeapAlloc( ghHeap, 0, count );
}


void __cdecl operator delete(
    void* object
    )
{
    HeapFree( ghHeap, 0, object );
}

extern "C" {
STDAPI DllGetClassObject(
    REFCLSID rclsid,
    REFIID riid,
    LPVOID * ppv
    )
{
    ND_ENTER( ND_DBG_NDI );

    UNREFERENCED_PARAMETER( rclsid );

    if( IsEqualIID( riid, IID_IClassFactory ) )
    {
        NetworkDirect::CClassFactory* pFactory = new NetworkDirect::CClassFactory();
        if( pFactory == NULL )
            return E_OUTOFMEMORY;

        *ppv = pFactory;
        return S_OK;
    }

    return E_NOINTERFACE;
}

STDAPI DllCanUnloadNow(void)
{
    ND_ENTER( ND_DBG_NDI );

    if( InterlockedCompareExchange( &NetworkDirect::gnRef, 0, 0 ) != 0 )
        return S_FALSE;

    return S_OK;
}

int
WSPAPI
WSPStartup(
    IN WORD wVersionRequested,
    OUT LPWSPDATA lpWSPData,
    IN LPWSAPROTOCOL_INFOW lpProtocolInfo,
    IN WSPUPCALLTABLE UpcallTable,
    OUT LPWSPPROC_TABLE lpProcTable
    )
{
    UNREFERENCED_PARAMETER( wVersionRequested );
    UNREFERENCED_PARAMETER( lpWSPData );
    UNREFERENCED_PARAMETER( lpProtocolInfo );
    UNREFERENCED_PARAMETER( UpcallTable );
    UNREFERENCED_PARAMETER( lpProcTable );
    return WSASYSNOTREADY;
}

static BOOL
_DllMain(
    IN                HINSTANCE                    hinstDll,
    IN                DWORD                        dwReason,
    IN                LPVOID                        lpvReserved )
{

    ND_ENTER( ND_DBG_NDI );

    UNUSED_PARAM( hinstDll );
    UNUSED_PARAM( lpvReserved );

    switch( dwReason )
    {
    case DLL_PROCESS_ATTACH:
        TCHAR    env_var[16];
        DWORD    i;


#if defined(EVENT_TRACING)
#if DBG
        WPP_INIT_TRACING(L"ibndprov.dll");
#else
        WPP_INIT_TRACING(L"ibndprov.dll");
#endif
#elif DBG 
        i = GetEnvironmentVariable( "IBNDPROV_DBG_LEVEL", env_var, sizeof(env_var) );
        if( i && i <= 16 )
        {
            g_nd_dbg_level = _tcstoul( env_var, NULL, 16 );
        }

        i = GetEnvironmentVariable( "IBNDPROV_DBG_FLAGS", env_var, sizeof(env_var) );
        if( i && i <= 16 )
        {
            g_nd_dbg_flags = _tcstoul( env_var, NULL, 16 );
        }

        if( g_nd_dbg_flags & ND_DBG_ERR )
            g_nd_dbg_flags |= CL_DBG_ERROR;

        ND_PRINT( TRACE_LEVEL_ERROR, ND_DBG_ERR ,
            ("(pcs %#x) IbNdProv: Debug print: level:%d, flags 0x%x\n",
            GetCurrentProcessId(), g_nd_dbg_level ,g_nd_dbg_flags) );
#endif

		i = GetEnvironmentVariable( "IBNDPROV_MAX_INLINE_SIZE", env_var, sizeof(env_var) );
		if( i && i <= 16 )
		{
			g_nd_max_inline_size = _tcstoul( env_var, NULL, 16 );
		}

        ghHeap = HeapCreate( 0, 0, 0 );
        if( ghHeap == NULL )
        {
            ND_PRINT_EXIT(
                TRACE_LEVEL_ERROR, ND_DBG_NDI, ("Failed to allocate private heap.\n") );
#if defined(EVENT_TRACING)
            WPP_CLEANUP();
#endif
            return FALSE;
        }

        ND_PRINT(TRACE_LEVEL_INFORMATION, ND_DBG_NDI, ("DllMain: DLL_PROCESS_ATTACH\n") );
        break;

    case DLL_PROCESS_DETACH:
        HeapDestroy( ghHeap );
        ND_PRINT(TRACE_LEVEL_INFORMATION, ND_DBG_NDI,
            ("DllMain: DLL_PROCESS_DETACH, ref count %d\n", NetworkDirect::gnRef) );

#if defined(EVENT_TRACING)
        WPP_CLEANUP();
#endif
        break;
    }

    ND_EXIT( ND_DBG_NDI );

    return TRUE;
}


extern BOOL APIENTRY
_DllMainCRTStartupForGS(
    IN                HINSTANCE                    h_module,
    IN                DWORD                        ul_reason_for_call, 
    IN                LPVOID                        lp_reserved );


BOOL APIENTRY
DllMain(
    IN                HINSTANCE                    h_module,
    IN                DWORD                        ul_reason_for_call, 
    IN                LPVOID                        lp_reserved )
{
    switch( ul_reason_for_call )
    {
    case DLL_PROCESS_ATTACH:
        if( !_DllMainCRTStartupForGS(
            h_module, ul_reason_for_call, lp_reserved ) )
        {
            return FALSE;
        }

        return _DllMain( h_module, ul_reason_for_call, lp_reserved );

    case DLL_THREAD_ATTACH:
        ND_PRINT(TRACE_LEVEL_INFORMATION, ND_DBG_NDI, ("DllMain: DLL_THREAD_ATTACH\n") );
        break;

    case DLL_THREAD_DETACH:
        ND_PRINT(TRACE_LEVEL_INFORMATION, ND_DBG_NDI, ("DllMain: DLL_THREAD_DETACH\n") );
        break;

    case DLL_PROCESS_DETACH:
        _DllMain( h_module, ul_reason_for_call, lp_reserved );

        return _DllMainCRTStartupForGS(
            h_module, ul_reason_for_call, lp_reserved );
    }
    return TRUE;
}

}   // extern "C"

